from ._stopping_criterion import StoppingCriterion
from ..accumulate_data import MLQMCData
from ..discrete_distribution import DigitalNetB2,Lattice,Halton
from ..discrete_distribution._discrete_distribution import LD
from ..true_measure import Gaussian
from ..integrand import MLCallOptions
from ..util import MaxSamplesWarning, ParameterError, MaxLevelsWarning
from numpy import *
from scipy.stats import norm
from time import time
import warnings


class CubQMCML(StoppingCriterion):
    """
    Stopping criterion based on multi-level quasi-Monte Carlo.

    >>> mlco = MLCallOptions(Lattice(seed=7))
    >>> sc = CubQMCML(mlco,abs_tol=.05)
    >>> solution,data = sc.integrate()
    >>> data
    MLQMCData (AccumulateData Object)
        solution        10.432
        n_total         172032
        n_level         [4096.  256.  256.  256.  256.  256.]
        levels          6
        mean_level      [10.053  0.183  0.102  0.054  0.027  0.014]
        var_level       [5.699e-05 5.129e-05 3.243e-05 1.610e-05 5.633e-06 1.108e-06]
        bias_estimate   0.007
        time_integrate  ...
    CubQMCML (StoppingCriterion Object)
        rmse_tol        0.019
        n_init          2^(8)
        n_max           10000000000
        replications    2^(5)
    MLCallOptions (Integrand Object)
        option          european
        sigma           0.200
        k               100
        r               0.050
        t               1
        b               85
        level           0
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      1
        decomp_type     PCA
    Lattice (DiscreteDistribution Object)
        d               1
        dvec            0
        randomize       SHIFT
        order           NATURAL
        gen_vec         1
        entropy         7
        spawn_key       ()

    References:
        
        [1] M.B. Giles and B.J. Waterhouse. 'Multilevel quasi-Monte Carlo path simulation'.
        pp.165-181 in Advanced Financial Modelling, in Radon Series on Computational and Applied Mathematics,
        de Gruyter, 2009. http://people.maths.ox.ac.uk/~gilesm/files/radon.pdf
    """

    def __init__(self, integrand, abs_tol=.05, alpha=.01, rmse_tol=None, n_init=256., n_max=1e10, 
        replications=32., levels_min=2, levels_max=10):
        """
        Args:
            integrand (Integrand): integrand with multi-level g method
            abs_tol (float): absolute tolerance
            alpha (float): uncertainty level.
                If rmse_tol not supplied, then rmse_tol = abs_tol/norm.ppf(1-alpha/2)
            rmse_tol (float): root mean squared error
                If supplied (not None), then absolute tolerance and alpha are ignored
                in favor of the rmse tolerance
            n_max (int): maximum number of samples
            replications (int): number of replications on each level
            levels_min (int): minimum level of refinement >= 2
            levels_max (int): maximum level of refinement >= Lmin
        """
        self.parameters = ['rmse_tol','n_init','n_max','replications']
        # initialization
        if rmse_tol:
            self.rmse_tol = float(rmse_tol)
        else: # use absolute tolerance
            self.rmse_tol =  float(abs_tol) / norm.ppf(1-alpha/2)
        self.n_init = float(n_init)
        self.n_max = float(n_max)
        self.replications = float(replications)
        self.levels_min = levels_min
        self.levels_max = levels_max
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        # Verify Compliant Construction
        allowed_levels = ['adaptive-multi']
        allowed_distribs = [LD]
        allow_vectorized_integrals = False
        super(CubQMCML,self).__init__(allowed_levels, allowed_distribs, allow_vectorized_integrals)

    def integrate(self):
        """ See abstract method. """
        # Construct AccumulateData Object to House Integration Data
        self.data = MLQMCData(self, self.integrand, self.true_measure, self.discrete_distrib,
            self.levels_min, self.levels_max, self.n_init, self.replications)
        t_start = time()
        while True:
            self.data.update_data()
            if self.data.var_level.sum() > (self.rmse_tol**2/2.):
                # double N_l on level with largest V_l/(2^l*N_l)
                efficient_level = argmax(self.data.var_cost_ratio_level)
                self.data.eval_level[efficient_level] = True
            elif self.data.bias_estimate > (self.rmse_tol/sqrt(2.)):
                if self.data.levels == self.levels_max + 1:
                        warnings.warn("Failed to achieve weak convergence. levels == levels_max.", MaxLevelsWarning)
                # add another level
                self.data._add_level()
            else:
                # both conditions met
                break
            total_next_samples = (self.data.replications*self.data.eval_level*self.data.n_level*2).sum()
            if (self.data.n_total + total_next_samples) > self.n_max:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples, which would exceed n_max = %d.
                Stopping integration process.
                Note that error tolerances may no longer be satisfied""" \
                % (int(self.data.n_total), int(total_next_samples), int(self.n_max))
                warnings.warn(warning_s, MaxSamplesWarning)
                break
        self.data.time_integrate = time() - t_start
        return self.data.solution,self.data
    
    def set_tolerance(self, abs_tol=None, alpha=.01, rmse_tol=None):
        """
        See abstract method. 
        
        Args:
            integrand (Integrand): integrand with multi-level g method
            abs_tol (float): absolute tolerance. Reset if supplied, ignored if not. 
            alpha (float): uncertainty level.
                If rmse_tol not supplied, then rmse_tol = abs_tol/norm.ppf(1-alpha/2)
            rel_tol (float): relative tolerance. Reset if supplied, ignored if not.
                Takes priority over absolute tolerance and alpha if supplied.
        """
        if rmse_tol != None:
            self.rmse_tol = float(rmse_tol)
        elif abs_tol != None:
            self.rmse_tol = (float(abs_tol) / norm.ppf(1-alpha/2.))



from .util import (
    bernoulli_poly,
    weighted_walsh_funcs)
from .fast_transforms import (
    fftbr,ifftbr,fwht
)
try:
    from .fast_transforms import (
        fftbr_torch,ifftbr_torch,fwht_torch
    )
except:
    pass
from .kernel import (
    KernelGaussian,
    KernelShiftInvar,KernelDigShiftInvar)
from .gram_matrix import (
    GramMatrix,
    FastGramMatrixLattice,FastGramMatrixDigitalNetB2)



from ._discrete_distribution import DiscreteDistribution
from .iid_std_uniform import IIDStdUniform
from .lattice import Lattice
from .digital_net_b2 import DigitalNetB2, Sobol, SobolSciPy
from .halton import Halton



import numpy as np 

def k4sumterm(x, t, cutoff=1e-8):
    r""" 
    $$\sum_{a=0}**{t-1} x_a / 2**{3a}$$
    where $x_a$ is the bit at index $a$ in the binary expansion of $x$
    e.g. $x = 6$ with $t=3$ has $(x_0,x_1,x_2) = (1,1,0)

    Args:
        x (np.ndarray or torch.Tensor): integer arrays
        t (int): number of bits in each integer
    
    >>> t = 3
    >>> x = np.random.randint(0,2**t,(5,4))
    >>> k4sumterm(x,t).shape
    (5, 4)
    >>> import torch 
    >>> xt = torch.randint(0,2**t,(5,4))
    >>> k4sumterm(xt,t).shape
    torch.Size([5, 4])
    """
    total = 0.
    for a in range(0,t):
        factor = 1/float(2**(3*a))
        if factor<cutoff: break
        total += (-1)**((x>>(t-a-1))&1)*factor
    return total

WEIGHTEDWALSHFUNCSPOS = {
    2: lambda beta,xf,xb,t: -beta*xf + 5/2*(1-2**(-beta)),
    3: lambda beta,xf,xb,t: beta*xf**2 - 5*(1-2**(-beta))*xf + 43/18*(1-2**(-2*beta)),
    4: lambda beta,xf,xb,t: -2/3*beta*xf**3 + 5*(1-2**(-beta))*xf**2 - 43/9*(1-2**(-2*beta))*xf + 701/294*(1-2**(-3*beta)) + beta*(1/48*k4sumterm(xb,t)-1/42),
}

WEIGHTEDWALSHFUNCSZEROS = {
    2: 5/2,
    3: 43/18,
    4: 701/294,
}

def weighted_walsh_funcs(alpha, xb, t):
    r"""
    Weighted walsh functions 
    $$\sum_{k=0}^\infty \phi_k(x) 2^{-\mu_\alpha(k)}$$ 
    where $\phi_k$ is the $k^\text{th}$ Walsh function 
    and $\mu_\alpha$ is the Dick weight function which sums the first $\alpha$ largest indices of 1 bits in the binary expansion of $k$ 
    e.g. $k=13=1101_2$ 1-bit indexes (4,3,1)$ so $\mu_1(k) = 4, \mu_2(k) = 4+3, \mu_3(k) = 4+3+1 = \mu_4(k) = \mu_5(k) = \dots$

    References: 
        
        [1] Dick, Josef. "Walsh spaces containing smooth functions and quasi–Monte Carlo rules of arbitrary high order." SIAM Journal on Numerical Analysis 46.3 (2008): 1519-1553.

        [2] Dick, Josef. "The decay of the Walsh coefficients of smooth functions." Bulletin of the Australian Mathematical Society 80.3 (2009): 430-453.

    Args:
        alpha (int): weighted walsh functions order 
        xb (np.ndarray or torch.Tensor): integer points at which to evaluate the weighted Walsh function 
        t (int): number of bits in each integer in xb
    
    >>> import torch 
    >>> t = 3 
    >>> xb = np.random.randint(0,2**t,(4,5))
    >>> xbt = torch.from_numpy(xb)
    >>> for alpha in list(WEIGHTEDWALSHFUNCSPOS.keys()):
    ...     y = weighted_walsh_funcs(alpha,xb,t)
    ...     yt = weighted_walsh_funcs(alpha,xbt,t)
    """
    assert isinstance(alpha,int)
    assert alpha in WEIGHTEDWALSHFUNCSPOS, "alpha = %d not in WEIGHTEDWALSHFUNCSPOS"%alpha
    assert alpha in WEIGHTEDWALSHFUNCSZEROS, "alpha = %d not in WEIGHTEDWALSHFUNCSZEROS"%alpha
    if isinstance(xb,np.ndarray):
        np_or_torch = np 
        y = np.ones(xb.shape) 
    else:
        import torch 
        np_or_torch = torch 
        y = torch.ones(xb.shape,device=xb.device)
    pidxs = xb>0
    y[~pidxs] = WEIGHTEDWALSHFUNCSZEROS[alpha]
    xfpidxs = (2**(-t))*xb[pidxs]
    betapidxs = -np_or_torch.floor(np_or_torch.log2(xfpidxs))
    y[pidxs] = WEIGHTEDWALSHFUNCSPOS[alpha](betapidxs,xfpidxs,xb[pidxs],t)
    return y



from ._integrand import Integrand
from .asian_option import AsianOption
from .european_option import EuropeanOption
from .keister import Keister
from .linear0 import Linear0
from .custom_fun import CustomFun
from .ml_call_options import MLCallOptions
from .box_integral import BoxIntegral
from .sensitivity_indices import SobolIndices,SensitivityIndices
from .ishigami import Ishigami
from .bayesian_lr_coeffs import BayesianLRCoeffs
from .um_bridge_wrapper import UMBridgeWrapper
from .genz import Genz
from .sin1d import Sin1d
from .hartmann6d import Hartmann6d
from .fourbranch2d import FourBranch2d
from .multimodal2d import Multimodal2d
from .barrier_option import BarrierOption


from .discrete_distribution import *
from .true_measure import *
from .integrand import *
from .stopping_criterion import *
from .kernel_methods import (
    fftbr,ifftbr,fwht,
    KernelShiftInvar,KernelDigShiftInvar,KernelGaussian,
    FastGramMatrixLattice,FastGramMatrixDigitalNetB2,GramMatrix)
try:
    from .kernel_methods import (
        fftbr_torch,ifftbr_torch,fwht_torch
    )
except:
    pass
from .util import plot_proj

name = "qmcpy"
__version__ = "1.6.2"



from ._true_measure import TrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution._discrete_distribution import DiscreteDistribution
from ..discrete_distribution import DigitalNetB2
from numpy import *
import scipy.stats


class SciPyWrapper(TrueMeasure):
    """
    Multivariate True Measure from Independent SciPy 1 Dimensional Marginals

    >>> unif_gauss_gamma = SciPyWrapper(
    ...     discrete_distrib = DigitalNetB2(3,seed=7),
    ...     scipy_distribs = [
    ...         scipy.stats.uniform(loc=1,scale=2),
    ...         scipy.stats.norm(loc=3,scale=4),
    ...         scipy.stats.gamma(a=5,loc=6,scale=7)])
    >>> unif_gauss_gamma.range
    array([[  1.,   3.],
           [-inf,  inf],
           [  6.,  inf]])
    >>> unif_gauss_gamma.gen_samples(4)
    array([[ 2.53470309,  0.35273007, 56.37744157],
           [ 1.35215166,  7.42509911, 33.50121705],
           [ 2.04111416,  3.83592672, 28.72235698],
           [ 1.8447669 , -4.11180523, 48.44339757]])
    >>> betas_2d = SciPyWrapper(discrete_distrib=DigitalNetB2(2,seed=7),scipy_distribs=scipy.stats.beta(a=5,b=1))
    >>> betas_2d.gen_samples(4)
    array([[0.59010522, 0.60012506],
           [0.9597241 , 0.9427888 ],
           [0.79325286, 0.98608598],
           [0.89417617, 0.76694864]])
    """
    
    def __init__(self, discrete_distrib, scipy_distribs):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform 
            scipy_distribs (list): instantiated CONTINUOUS UNIVARIATE scipy.stats distributions 
                https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions
        """
        self.domain = array([[0,1]])
        if not isinstance(discrete_distrib,DiscreteDistribution):
            raise ParameterError("SciPyWrapper requires discrete_distrib be a DiscreteDistribution.")
        self._parse_sampler(discrete_distrib)
        self.scipy_distrib = list(scipy_distribs) if not isinstance(scipy_distribs,scipy.stats._distn_infrastructure.rv_continuous_frozen) else [scipy_distribs]
        for sd in self.scipy_distrib:
            if isinstance(sd,scipy.stats._distn_infrastructure.rv_continuous_frozen): continue
            raise ParameterError('''
                SciPyWrapper requires each value of scipy_distribs to be a 
                1 dimensional scipy.stats continuous distribution, 
                see https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions.''')
        self.sds = self.scipy_distrib if len(self.scipy_distrib)>1 else self.scipy_distrib*discrete_distrib.d
        if len(self.sds)!=discrete_distrib.d:
            raise DimensionError("length of scipy_distribs must match the dimension of the discrete_distrib")
        self.range = array([sd.interval(1) for sd in self.sds])
        super(SciPyWrapper,self).__init__()

    def _transform(self, x):
        return array([self.sds[j].ppf(x[:,j]) for j in range(self.d)],dtype=float).T
    
    def _weight(self, x):
        return array([self.sds[j].pdf(x[:,j]) for j in range(self.d)],dtype=float).T.prod(1)
    
    def _spawn(self, sampler, dimension):
        return SciPyWrapper(sampler,self.scipy_distrib)
    


from ._integrand import Integrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Gaussian
from numpy import *
from scipy.special import gamma


class Keister(Integrand):
    """
    $f(\\boldsymbol{t}) = \\pi^{d/2} \\cos(\\| \\boldsymbol{t} \\|)$.

    The standard example integrates the Keister integrand with respect to an 
    IID Gaussian distribution with variance 1./2.

    >>> k = Keister(DigitalNetB2(2,seed=7))
    >>> x = k.discrete_distrib.gen_samples(2**10)
    >>> y = k.f(x)
    >>> y.mean()
    1.808...
    >>> k.true_measure
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      2^(-1)
        decomp_type     PCA
    >>> k = Keister(Gaussian(DigitalNetB2(2,seed=7),mean=0,covariance=2))
    >>> x = k.discrete_distrib.gen_samples(2**12)
    >>> y = k.f(x)
    >>> y.mean()
    1.808...
    >>> yp = k.f(x,periodization_transform='c2sin')
    >>> yp.mean()
    1.807...

    References:

        [1] B. D. Keister, Multidimensional Quadrature Algorithms, 
        `Computers in Physics`, *10*, pp. 119-122, 1996.
    """

    def __init__(self, sampler):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform
        """
        self.sampler = sampler
        self.true_measure = Gaussian(self.sampler,mean=0,covariance=1/2)
        super(Keister,self).__init__(dimension_indv=1,dimension_comb=1,parallel=False)
    
    def g(self, t):
        d = t.shape[1]
        norm = sqrt((t**2).sum(1))
        k = pi**(d/2)*cos(norm)
        return k
    
    def _spawn(self, level, sampler):
        return Keister(sampler=sampler)

    def exact_integ(self, d):
        """
        computes the true value of the Keister integral in dimension d.
        Accuracy might degrade as d increases due to round-off error.
        :param d:
        :return: true_integral
        """
        cosinteg = zeros(shape=(d))
        cosinteg[0] = sqrt(pi) / (2 * exp(1 / 4))
        sininteg = zeros(shape=(d))
        sininteg[0] = 4.244363835020225e-01
        cosinteg[1] = (1 - sininteg[0]) / 2
        sininteg[1] = cosinteg[0] / 2
        for j in range(2, d):
            cosinteg[j] = ((j-1)*cosinteg[j-2]-sininteg[j-1])/2
            sininteg[j] = ((j-1)*sininteg[j-2]+cosinteg[j-1])/2
        I = (2*(pi**(d/2))/gamma(d/2))*cosinteg[d-1]
        return I



from ..discrete_distribution import DigitalNetB2
from ._integrand import Integrand
from ..true_measure import BrownianMotion
from ..util import ParameterError
from numpy import *

class BarrierOption(Integrand):
    """
    Barrier Option. 

    >>> barrier_option = BarrierOption(DigitalNetB2(4,seed=7))
    >>> barrier_option
    BarrierOption (Integrand Object)
        volatility      0.200
        call_put        call
        start_price     30
        strike_price    35
        barrier_price   38
        interest_rate   0.050
        in_out          in
        dim_frac        0
    >>> x = barrier_option.discrete_distrib.gen_samples(2**12)
    >>> y = barrier_option.f(x)
    >>> y.mean()
    1.08...
    >>> level_dims = [2,4,8]
    >>> barrier_option_multilevel = BarrierOption(DigitalNetB2(seed=7),multilevel_dims=level_dims)
    >>> levels_to_spawn = arange(barrier_option_multilevel.max_level+1)
    >>> barrier_option_single_levels = barrier_option_multilevel.spawn(levels_to_spawn)
    >>> yml = 0
    >>> for barrier_option_single_level in barrier_option_single_levels:
    ...     x = barrier_option_single_level.discrete_distrib.gen_samples(2**12)
    ...     level_est = barrier_option_single_level.f(x).mean()
    ...     yml += level_est
    >>> yml
    1.10...
    """

    def __init__(self, sampler, volatility=0.2, start_price=30., strike_price=35., barrier_price = 38.,\
        interest_rate=0.05, t_final=1., call_put='call',in_out = 'in',multilevel_dims = None, decomp_type='PCA', _dim_frac=0):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform
            volatility (float): sigma, the volatility of the asset
            start_price (float): S(0), the asset value at t=0
            strike_price (float): strike_price, the call/put offer
            barrier_price (float): price at which the option activates/deactivates
            interest_rate (float): r, the annual interest rate
            t_final (float): exercise time
            multilevel_dims (list of ints): list of dimensions at each level. 
                Leave as None for single-level problems
            _dim_frac (float): for internal use only, users should not set this parameter. 
        """

        self.parameters = ['volatility', 'call_put', 'start_price', 'strike_price', 'barrier_price', 'interest_rate','in_out']
        self.sampler = sampler
        self.true_measure = BrownianMotion(self.sampler,t_final,decomp_type=decomp_type)
        self.volatility = float(volatility)
        if(barrier_price == start_price):
            raise ParameterError("Barrier Price must be greater or less than the start price. They can't be equal.")
        self.start_price = float(start_price)
        self.strike_price = float(strike_price)
        self.barrier_price = float(barrier_price)
        self.interest_rate = float(interest_rate)
        self.t_final = t_final
        self.decomp_type = decomp_type
        self.call_put = call_put.lower()
        if self.call_put not in ['call','put']:
            raise ParameterError("call_put must be either 'call' or 'put'")
        self.in_out = in_out.lower()
        if self.in_out not in ['in', 'out']:
            raise ParameterError("in_out must be either 'in' or 'out'")
        if(self.start_price < self.barrier_price):
            self.up = True
        else:
            self.up = False
        if(self.in_out == 'in'):
            self.i = True
        else:
            self.i = False
        if(self.call_put == 'call'):
            self.call = True
        else:
            self.call = False
        # handle single vs multilevel
        self.multilevel_dims = multilevel_dims
        if self.multilevel_dims is not None: # multi-level problem
            self.dim_fracs = array(
                [0]+ [float(self.multilevel_dims[i])/float(self.multilevel_dims[i-1]) 
                for i in range(1,len(self.multilevel_dims))],
                dtype=float)
            self.max_level = len(self.multilevel_dims)-1
            self.leveltype = 'fixed-multi'
            self.parent = True
            self.parameters += ['multilevel_dims']
        else: # single level problem
            self.dim_frac = _dim_frac
            self.leveltype = 'single'
            self.parent = False
            self.parameters += ['dim_frac']
        super(BarrierOption,self).__init__(dimension_indv=1,dimension_comb=1,parallel=False)   

    def _get_discounted_payoffs(self,stock_path,dimension):
        """
        Calculate the discounted payoff from the stock path. 
        
        Args:
            stock_path (ndarray): n samples by d dimension option prices at monitoring times
            dimension (int): number of dimensions
        
        Return:
            ndarray: n vector of discounted payoffs
        """
        expected_stock = stock_path[:,dimension - 1]
        if(self.call):
            disc_payoff = expected_stock - self.strike_price
        else:
            disc_payoff = self.strike_price - expected_stock
        if(self.up and self.i):
            bar_flag = stock_path >= self.barrier_price
            bar_flag = bar_flag.sum(axis = 1) > 0
        elif(self.up and (self.i is False)):
            bar_flag = stock_path < self.barrier_price
            bar_flag = bar_flag.sum(axis = 1) == dimension
        elif((self.up is False) and self.i):
            bar_flag = stock_path <= self.barrier_price
            bar_flag = bar_flag.sum(axis = 1) > 0
        else:
            bar_flag = stock_path > self.barrier_price
            bar_flag = bar_flag.sum(axis = 1) == dimension
        disc_payoff = disc_payoff*bar_flag
        disc_payoff = maximum(zeros(disc_payoff.size), disc_payoff)
        return (disc_payoff * exp(-self.interest_rate * self.t_final))
    
    def g(self, t):
        if self.parent:
            raise ParameterError('''
                Cannot evaluate an integrand with multilevel_dims directly,
                instead spawn some children and evaluate those.''')
        self.s_fine = self.start_price * exp(
            (self.interest_rate - self.volatility ** 2 / 2.) *
            self.true_measure.time_vec + self.volatility * t)
        for xx,yy in zip(*where(self.s_fine<0)): # if stock becomes <=0, 0 out rest of path
            self.s_fine[xx,yy:] = 0
        y = self._get_discounted_payoffs(self.s_fine,self.d)
        if self.dim_frac > 0:
            s_course = self.s_fine[:, int(self.dim_frac - 1):: int(self.dim_frac)]
            d_course = int(float(self.d) / self.dim_frac)
            y_course = self._get_discounted_payoffs(s_course, d_course)
            y -= y_course
        return y
    
    def _dimension_at_level(self, level):
        return self.d if self.multilevel_dims is None else self.multilevel_dims[level]
        
    def _spawn(self, level, sampler):            
        return BarrierOption(
            sampler = sampler,
            volatility = self.volatility,
            start_price = self.start_price,
            strike_price = self.strike_price,
            barrier_price = self.barrier_price,
            interest_rate = self.interest_rate,
            t_final = self.t_final,
            call_put = self.call_put,
            in_out = self.in_out,
            _dim_frac = self.dim_fracs[level] if hasattr(self,'dim_fracs') else 0)
            



from ._true_measure import TrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution import DigitalNetB2
from numpy import *
from scipy.stats import norm


class JohnsonsSU(TrueMeasure):
    """
    >>> jsu = JohnsonsSU(DigitalNetB2(2,seed=7),gamma=1,xi=2,delta=3,lam=4)
    >>> jsu.gen_samples(4)
    array([[-1.66790878, -1.58770999],
           [ 1.85795234,  1.54353235],
           [-0.06077722,  2.66089215],
           [ 0.89373077, -0.27712026]])
    >>> jsu
    JohnsonsSU (TrueMeasure Object)
        gamma           1
        xi              2^(1)
        delta           3
        lam             2^(2)
    
    See https://en.wikipedia.org/wiki/Johnson%27s_SU-distribution
    """

    def __init__(self, sampler, gamma=1, xi=1, delta=2, lam=2):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform 
            gamma (ndarray): gamma
            xi (ndarray): xi
            delta (ndarray): delta > 0
            lam (ndarray): lambda > 0
        """
        self.parameters = ['gamma', 'xi', 'delta', 'lam']
        self.domain = array([[0,1]])
        self.range = array([[-inf,inf]])
        self._parse_sampler(sampler)
        self.gamma = gamma
        self.xi = xi
        self.delta = delta
        self.lam = lam
        if isscalar(self.gamma):
            gamma = tile(self.gamma,self.d)
        if isscalar(self.xi):
            xi = tile(self.xi,self.d)
        if isscalar(self.delta):
            delta = tile(self.delta,self.d)
        if isscalar(self.lam):
            lam = tile(self.lam,self.d)
        self._gamma = array(gamma)
        self._xi = array(xi)
        self._delta = array(delta)
        self._lam = array(lam)
        if len(self._gamma)!=self.d or len(self._xi)!=self.d or len(self._delta)!=self.d or len(self._lam)!=self.d:
            raise DimensionError("all Johnson's S_U parameters be scalar or have length equal to dimension.")
        if (self._delta<=0).any() or (self._lam<=0).any():
            raise ParameterError("delta and lam (lambda) must be greater than 0")
        super(JohnsonsSU,self).__init__() 

    def _transform(self, x):
        return self._lam*sinh((norm.ppf(x)-self._gamma)/self._delta)+self._xi
    
    def _weight(self, x):
        term1 = (x-self._xi)/self._lam
        term2 = self._delta/(self._lam*sqrt(2*pi)) * 1/sqrt(1+term1**2)
        term3 = exp(-1/2*(self._gamma+self._delta*arcsinh(term1))**2)
        return prod( term2*term3, 1)
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = JohnsonsSU(sampler,gamma=self._gamma,xi=self._xi,delta=self._delta,lam=self._lam)
        else:
            gamma = self._gamma[0]
            xi = self._xi[0]
            delta = self._delta[0]
            lam = self._lam[0]
            if not ( all(self._gamma==gamma) and all(self._xi==xi) and all(self._delta==delta) and all(self._lam==lam) ):
                raise DimensionError('''
                    In order to spawn a JohnsonsSU measure
                    gamma must all be the same and 
                    xi must all be the same and 
                    delta must all be the same and 
                    lam (lambda) must all be the same.''')
            spawn = JohnsonsSU(sampler,gamma=gamma,xi=xi,delta=delta,lam=lam)
        return spawn



from ._kernel_prod import _KernelProd
from ..util import bernoulli_poly,weighted_walsh_funcs
from ...discrete_distribution import Lattice,DigitalNetB2
import numpy as np

class _KernelProdSpecial(_KernelProd):
    def parsed_call(self, x1, x2, n1, n2, beta1s, beta2s, m1, m2, c1s, c2s, diag_only):
        if diag_only:
            delta = self.x1_ominus_x2(x1[:,None,:],x2[:,None,:])
            assert delta.ndim==3 and delta.shape[2]==self.d
            inds,idxs,consts = self.inds_idxs_consts(beta1s[:,None,:],beta2s[None,:,:])
            k = self.npt.ones((m1,1,n1,1,self.d),dtype=float,**self.ckwargs)
            self.eval_low_low(k,delta,inds,idxs,consts,m1,1)
        else:
            delta = self.x1_ominus_x2(x1[:,None,:],x2[None,:,:]) # n1 x n2 x d
            assert delta.ndim==3 and delta.shape[2]==self.d
            inds,idxs,consts = self.inds_idxs_consts(beta1s[:,None,:],beta2s[None,:,:]) # m1 x m2 x n1 x n2 x d
            k = self.npt.ones((m1,m2,n1,n2,self.d),dtype=float,**self.ckwargs)
            self.eval_low_low(k,delta,inds,idxs,consts,m1,m2) # m1 x m2 x n1 x n2 x d
        kcomb = (inds[:,:,None,None,:]+self.lengthscales*k).prod(-1) # m1 x m2 x n1 x n2
        kcomb = (c1s[:,None,None,None]*c2s[None,:,None,None]*kcomb).sum((0,1)) # n1 x n2
        kmat = self.scale*kcomb
        return kmat[:,0] if diag_only else kmat
    def eval_low_u_noscale(self, u, delta_u, inds, idxs, consts, m1, m2, n1, n2, d_u):
        k_u = self.npt.ones((m1,m2,n1,n2,d_u),dtype=float,**self.ckwargs)
        self.eval_low_low(k_u,delta_u,inds[:,:,u],idxs[:,:,u],consts[:,:,u],m1,m2)
        kcomb_u = (inds[:,:,None,None,u]+self.lengthscales[u]*k_u).prod(-1) # m1 x m1 x n1 x n2
        return kcomb_u

class KernelShiftInvar(_KernelProdSpecial):
    """
    Shift invariant kernel with product weights and derivatives 

    >>> d = 3
    >>> x = Lattice(d,randomize="shift",order="linear",seed=7).gen_samples(8)
    >>> k = KernelShiftInvar(d,torchify=False)
    >>> with np.printoptions(precision=2):
    ...     k(x,x)
    array([[27.22,  0.41,  0.98, -2.41, -0.98, -2.41,  0.98,  0.41],
           [ 0.41, 27.22,  0.41,  0.98, -2.41, -0.98, -2.41,  0.98],
           [ 0.98,  0.41, 27.22,  0.41,  0.98, -2.41, -0.98, -2.41],
           [-2.41,  0.98,  0.41, 27.22,  0.41,  0.98, -2.41, -0.98],
           [-0.98, -2.41,  0.98,  0.41, 27.22,  0.41,  0.98, -2.41],
           [-2.41, -0.98, -2.41,  0.98,  0.41, 27.22,  0.41,  0.98],
           [ 0.98, -2.41, -0.98, -2.41,  0.98,  0.41, 27.22,  0.41],
           [ 0.41,  0.98, -2.41, -0.98, -2.41,  0.98,  0.41, 27.22]])
    >>> import torch 
    >>> xt = torch.from_numpy(x).float()
    >>> kt = KernelShiftInvar(d,torchify=True)
    >>> kt(xt,xt[[0]])
    tensor([[27.2208],
            [ 0.4138],
            [ 0.9768],
            [-2.4126],
            [-0.9776],
            [-2.4126],
            [ 0.9768],
            [ 0.4138]], dtype=torch.float64)
    >>> kt(xt,xt,diag_only=True)
    tensor([27.2208, 27.2208, 27.2208, 27.2208, 27.2208, 27.2208, 27.2208, 27.2208],
           dtype=torch.float64)
    """
    DEFAULTALPHA = 4
    def _my_init(self, **kwargs):
        if self.torchify:
            self.lgamma = self.npt.lgamma
        else:
            import scipy.special
            self.lgamma = scipy.special.loggamma
    def inds_idxs_consts(self, beta1, beta2):
        beta1pbeta2 = beta1+beta2
        inds = 1.*(beta1pbeta2==0)
        idxs = 2*self.alpha-beta1pbeta2
        consts = (-1)**(self.alpha+beta2+1)*self.npt.exp(2*self.alpha*np.log(2*np.pi)-self.lgamma(idxs+1))
        return inds,idxs,consts
    def x1_ominus_x2(self, x1, x2):
        delta = (x1-x2)%1
        return delta        
    def eval_low_low(self, k, delta, inds, idxs, consts, m1, m2):
        for l1 in range(m1):
            for l2 in range(m2):
                for j in range(delta.shape[2]):
                    k[l1,l2,:,:,j] = consts[l1,l2,j]*bernoulli_poly(idxs[l1,l2,j].item(),delta[:,:,j])
    
class KernelDigShiftInvar(_KernelProdSpecial):
    """
    Digitally Shift invariant kernel with product weights and derivatives 
    
    >>> d = 3
    >>> dnb2 = DigitalNetB2(d,randomize="LMS_DS",seed=7)
    >>> xb = dnb2.gen_samples(8,return_binary=True)
    >>> k = KernelDigShiftInvar(d,t=dnb2.t_lms,torchify=False)
    >>> with np.printoptions(precision=2):
    ...     k(xb,xb)
    array([[13.63,  0.17,  0.21,  0.86,  1.17,  0.28,  0.45,  0.43],
           [ 0.17, 13.63,  0.86,  0.21,  0.28,  1.17,  0.43,  0.45],
           [ 0.21,  0.86, 13.63,  0.17,  0.45,  0.43,  1.17,  0.28],
           [ 0.86,  0.21,  0.17, 13.63,  0.43,  0.45,  0.28,  1.17],
           [ 1.17,  0.28,  0.45,  0.43, 13.63,  0.17,  0.21,  0.86],
           [ 0.28,  1.17,  0.43,  0.45,  0.17, 13.63,  0.86,  0.21],
           [ 0.45,  0.43,  1.17,  0.28,  0.21,  0.86, 13.63,  0.17],
           [ 0.43,  0.45,  0.28,  1.17,  0.86,  0.21,  0.17, 13.63]])
    >>> import torch
    >>> kt = KernelDigShiftInvar(d,torchify=True,t=dnb2.t_lms)
    >>> kt(xb,xb[[0]])
    tensor([[13.6329],
            [ 0.1689],
            [ 0.2086],
            [ 0.8560],
            [ 1.1687],
            [ 0.2796],
            [ 0.4520],
            [ 0.4332]], dtype=torch.float64)
    >>> kt(xb,xb,diag_only=True)
    tensor([13.6329, 13.6329, 13.6329, 13.6329, 13.6329, 13.6329, 13.6329, 13.6329],
           dtype=torch.float64)
    """
    DEFAULTALPHA = 3
    def _my_init(self, **kwargs):
        self.t = kwargs["t"] if "t" in kwargs else None
    def set_t(self, t):
        self.t = t
    def inds_idxs_consts(self, beta1, beta2):
        beta1pbeta2 = beta1+beta2
        inds = 1.*(beta1pbeta2==0)
        idxs = self.alpha-beta1pbeta2
        consts = (-2)**beta1pbeta2
        return inds,idxs,consts
    def x1_ominus_x2(self, x1, x2):
        delta = x1^x2
        return delta        
    def eval_low_low(self, k, delta, inds, idxs, consts, m1, m2):
        for l1 in range(m1):
            for l2 in range(m2):
                for j in range(delta.shape[2]):
                    k[l1,l2,:,:,j] = consts[l1,l2,j]*weighted_walsh_funcs(idxs[l1,l2,j].item(),delta[:,:,j],self.t)-inds[l1,l2,j]
                    


from ._gram_matrix import _GramMatrix
from ...discrete_distribution import IIDStdUniform
from ..kernel import KernelGaussian
import numpy as np
import itertools

class GramMatrix(_GramMatrix):
    """ Gram Matrix 

    >>> import torch 

    >>> d = 3
    >>> n = 2**3
    >>> dd_obj = IIDStdUniform(d,seed=7)
    >>> x = torch.from_numpy(dd_obj.gen_samples(n))
    >>> kernel_obj = KernelGaussian(d,torchify=True)
    >>> lbetas = [
    ...     torch.tensor([0,0,0]),
    ...     torch.tensor([0,2,0])]
    >>> gm = GramMatrix(kernel_obj,x,x,lbetas,lbetas)
    >>> gm._check()
    >>> y = torch.from_numpy(dd_obj.rng.uniform(size=gm.t1*n))
    >>> gm@y
    tensor([1.2413, 1.2839, 1.3816, 1.2028, 1.3405, 1.4576, 1.3655, 1.2973, 2.5923,
            1.9923, 1.9979, 2.3267, 2.0339, 1.4817, 1.4921, 1.9105],
           dtype=torch.float64)
    >>> v = gm.solve(y)
    >>> v
    tensor([-14.9060,  12.7191,  -5.8524,   9.7757,   6.9070,   8.6440,  -3.3528,
            -13.0816,  -5.6176,   2.3622,  15.8668,   5.2045,   2.1466, -11.1619,
             -1.8097,  -8.8931], dtype=torch.float64)
           
    >>> lbetas = [
    ...     torch.tensor([1,0,0]),
    ...     torch.tensor([[0,1,0],[0,0,1]]),
    ...     [torch.tensor([1,0,0]),torch.tensor([0,1,0])],
    ...     [torch.tensor([[1,0,0],[0,1,0],[0,0,1]]),torch.tensor([[1,0,1],[0,1,0],[0,0,0]])]
    ...     ]
    >>> for n1,n2 in itertools.product([n//2,n],[n//2,n]):
    ...     for ib1,ib2 in itertools.product(range(len(lbetas)),range(len(lbetas))):
    ...         x1 = x[:n1]
    ...         x2 = x[:n2]
    ...         lbeta1s = lbetas[ib1] 
    ...         lbeta2s = lbetas[ib2]
    ...         gm = GramMatrix(kernel_obj,x1,x2,lbeta1s,lbeta2s,adaptive_noise=False)
    ...         gm._check()
    """
    def __init__(self, kernel_obj, x1, x2, lbeta1s=0, lbeta2s=0, lc1s=1., lc2s=1., noise=1e-8, adaptive_noise=True):
        """
        Args:
            kernel_obj (kernel): kernel
            x1 (np.ndarray or torch.Tensor): n1 left locations
            x2 (np.ndarray or torch.Tensor): n2 right locations
            lbeta1s (list of either np.ndarray or torch.Tensor): list of (m1[l],d) arrays of first derivative orders
            lbeta1s (list of either np.ndarray or torch.Tensor): list of (m2[l],d) arrays of first derivative orders
            c1s (list of either np.ndarray or torch.Tensor): list of length m1[l] vectors of derivative coefficients 
            c2s (list of either np.ndarray or torch.Tensor): list of length m2[l] vector of derivative coefficients
            noise (float): nugget term 
            adaptive_noise (bool): if True, use the adative noise scheme described in Appendix A of 
                Chen, Yifan, et al. "Solving and learning nonlinear PDEs with Gaussian processes." 
                Journal of Computational Physics 447 (2021): 110668.
        """
        super(GramMatrix,self).__init__(kernel_obj,noise,lbeta1s,lbeta2s,lc1s,lc2s)
        self.x1,self.x2 = x1,x2
        self.n1 = len(self.x1) 
        self.n2 = len(self.x2) 
        assert self.x1.shape==(self.n1,self.d) and self.x2.shape==(self.n2,self.d)
        self.full_mat = self._construct_full_gram_matrix(self.x1,self.x2,self.t1,self.t2,self.lbeta1s,self.lbeta2s,self.lc1s_og,self.lc2s_og)
        self.size = (self.n1*self.t1,self.n2*self.t2)
        invertible_conds = [
            ( self.get_ptr(self.x1)==self.get_ptr(self.x2), "x1 and x2 must point to the same object"),
            ( self.n1==self.n2, "require square matrices"),
            ( self.t1==self.t2 and all((self.lbeta1s[tt1]==self.lbeta2s[tt1]).all() and (self.lc1s[tt1]==self.lc2s[tt1]).all() for tt1 in range(self.t1)), "require lbeta1s=lbeta2s and lc1s=lc2s"),
            ]  
        super(GramMatrix,self)._set_invertible_conds(invertible_conds)
        if self.invertible:
            if adaptive_noise:
                assert np.isscalar(noise) and noise>0
                assert (self.lbeta1s[0]==0.).all() and (self.lbeta1s[0].shape==(1,self.d)) and (self.lc1s_og[0]==1.).all() and (self.lc1s_og[0].shape==(1,))
                traces = self.full_mat.diagonal()[::self.n1]
                trace_ratios = traces/traces[0]
                self.full_mat += noise*self.npt.diag((self.npt.ones((self.n1,len(traces)))*trace_ratios).T.flatten())
            else:
                if np.isscalar(noise) or (noise.ndim==1 and len(noise)==1):
                    noise = noise*self.npt.ones(self.t1,dtype=float,**self.ckwargs)
                assert noise.ndim==1 and len(noise)==self.t1 and (noise>=0).all()
                noisy_diag = (noise[:,None]*self.npt.ones((1,self.n1),dtype=float,**self.ckwargs)).flatten()  
                size_range = self.npt.arange(self.size[0],dtype=int,**self.ckwargs)
                self.full_mat[size_range,size_range] = self.full_mat[size_range,size_range]+noisy_diag
        self._l_chol = None 
    @property
    def l_chol(self):
        if self._l_chol is None: 
            self._l_chol = self.cholesky(self.full_mat)
        return self._l_chol
    def get_new_left_full_gram_matrix(self, new_x, new_lbetas=0, new_lcs=1.):
        new_lbetas,new_lcs,new_t,new_m = self._parse_lbetas_lcs(new_lbetas,new_lcs)
        gm = self._construct_full_gram_matrix(self.x1,new_x,self.t1,new_t,self.lbeta1s,new_lbetas,self.lc1s_og,new_lcs).T
        return gm
    def _init_invertibile(self):
        self.l_chol = self.cholesky(self.full_mat)
    def logdet(self):
        assert self.invertible,"logdet currently only supports invertible matrices"
        if self.torchify:
            import torch 
            return torch.logdet(self.full_mat)
        else:
            sign,val = np.linalg.slogdet(self.full_mat)
            assert sign>0, "det <= 0, which should not happen for inverbible matrices"
            return val
    def multiply(self, *args, **kwargs):
        return self.__matmul__(*args, **kwargs)
    def __matmul__(self, y):
        return self.full_mat@y
    def solve(self, y):
        assert self.invertible, self.invertible_error_msg
        return self.cho_solve(self.l_chol,y)
    def _check(self):
        if not self.invertible: return
        rng = np.random.Generator(np.random.SFC64(7))
        y = rng.uniform(size=(self.n2*self.t2,2))
        if self.npt==np:
            assert np.allclose(self.solve(y[:,0]),np.linalg.solve(self.full_mat,y[:,0]),rtol=2.5e-2)
            assert np.allclose(self.solve(y),np.linalg.solve(self.full_mat,y),rtol=2.5e-2)
        else:
            import torch
            y = torch.from_numpy(y)
            assert np.allclose(self.solve(y[:,0]).numpy(),np.linalg.solve(self.full_mat.numpy(),y[:,0]),rtol=2.5e-2)
            assert np.allclose(self.solve(y).numpy(),np.linalg.solve(self.full_mat.numpy(),y),rtol=2.5e-2)



from ._integrand import Integrand
from . import Keister, BoxIntegral
from ..stopping_criterion import CubQMCNetG
from ..util import ParameterError
from ..true_measure import Uniform
from ..discrete_distribution import DigitalNetB2
from numpy import *
from itertools import combinations


class SensitivityIndices(Integrand):
    """
    Sensitivity' Indices, normalized Sobol' Indices. 

    >>> dnb2 = DigitalNetB2(dimension=3,seed=7)
    >>> keister_d = Keister(dnb2)
    >>> keister_indices = SensitivityIndices(keister_d,indices='singletons')
    >>> sc = CubQMCNetG(keister_indices,abs_tol=1e-3)
    >>> solution,data = sc.integrate()
    >>> solution.squeeze()
    array([[0.32822718, 0.32808433, 0.32795943],
           [0.33865673, 0.33869732, 0.33859364]])
    >>> data
    LDTransformData (AccumulateData Object)
        solution        [[0.328 0.328 0.328]
                        [0.339 0.339 0.339]]
        comb_bound_low  [[0.328 0.328 0.327]
                        [0.338 0.338 0.338]]
        comb_bound_high [[0.329 0.329 0.328]
                        [0.34  0.34  0.339]]
        comb_flags      [[ True  True  True]
                        [ True  True  True]]
        n_total         2^(16)
        n               [[[65536. 65536. 65536.]
                         [65536. 65536. 65536.]
                         [65536. 65536. 65536.]]
    <BLANKLINE>
                        [[32768. 32768. 32768.]
                         [32768. 32768. 32768.]
                         [32768. 32768. 32768.]]]
        time_integrate  ...
    CubQMCNetG (StoppingCriterion Object)
        abs_tol         0.001
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
    SensitivityIndices (Integrand Object)
        indices         [[0]
                        [1]
                        [2]]
        n_multiplier    3
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      2^(-1)
        decomp_type     PCA
    DigitalNetB2 (DiscreteDistribution Object)
        d               6
        dvec            [0 1 2 3 4 5]
        randomize       LMS_DS
        graycode        0
        entropy         7
        spawn_key       (0,)
    >>> sc = CubQMCNetG(SobolIndices(BoxIntegral(DigitalNetB2(3,seed=7)),indices='all'),abs_tol=.01)
    >>> sol,data = sc.integrate()
    >>> print(sol)
    [[[0.32392078 0.33342691]
      [0.32267736 0.33338337]
      [0.32301062 0.33337195]
      [0.65550663 0.66671246]
      [0.65536219 0.66671197]
      [0.65543028 0.66668495]]
    <BLANKLINE>
     [[0.34472618 0.33355823]
      [0.34473257 0.33336069]
      [0.34388133 0.3333552 ]
      [0.67725499 0.66676409]
      [0.67629333 0.66673255]
      [0.67729945 0.6667438 ]]]
    
    References: 
        [1] Art B. Owen.Monte Carlo theory, methods and examples. 2013. Appendix A.
    """
    def __init__(self, integrand, indices='singletons'):
        """
        Args:
            integrand (Integrand): integrand to find Sobol' indices of
            indices (list of lists): each element of indices should be a list of indices, u,
                at which to compute the Sobol' indices. 
                The default indices='singletons' sets indices=[[0],[1],...[d-1]]. 
                Should not include [], the null set, or [0,...,d-1], the set of all indices. 
                Setting indices='all' will compute all sensitivity indices
        """
        self.parameters = ['indices','n_multiplier']
        self.integrand = integrand
        self.d = self.integrand.d
        # indices
        self.indices = indices
        if self.indices=='singletons':
            self.indices = [[j] for j in range(self.d)]
        elif self.indices=='all':
            self.indices = []
            for r in range(1,self.d):
                self.indices += [list(idx) for idx in combinations(arange(self.d),r)]
        if [] in self.indices or [i for i in range(self.d)] in self.indices:
            raise ParameterError('SensitivityIndices indices cannot include [], the null set.')
        self.s = len(self.indices)
        self.indices_bool_mat = tile(False,(self.s,self.d))
        for k in range(self.s): self.indices_bool_mat[k,self.indices[k]] = True
        self.not_indices_bool_mat = ~self.indices_bool_mat
        # sensitivity_index
        self.n_multiplier = self.s
        self.dtilde = 2*self.d
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.true_measure.discrete_distrib.spawn(s=1,dimensions=[self.dtilde])[0]
        self.sampler = self.integrand.sampler
        super(SensitivityIndices,self).__init__(
            dimension_indv = (2,3,self.s,)+self.integrand.d_indv,
            dimension_comb = (2,self.s,)+self.integrand.d_indv,
            parallel = False)
    
    def f(self, x, *args, **kwargs):
        z = x[:,self.d:]
        x = x[:,:self.d]
        n,d = x.shape
        y = zeros((n,)+self.d_indv,dtype=float)
        compute_flags = kwargs['compute_flags']
        del kwargs['compute_flags']
        v = zeros((n,d),dtype=float)
        f_x = self.integrand.f(x,*args,**kwargs)
        f_z = self.integrand.f(z,*args,**kwargs)
        for k in range(self.s):
            flags_closed = compute_flags[0,0,k,:]
            flags_total = compute_flags[1,0,k,:]
            flags_k = flags_closed|flags_total
            if not flags_k.any(): continue
            u_bool = self.indices_bool_mat[k]
            not_u_bool = self.not_indices_bool_mat[k]
            v[:,u_bool] = x[:,u_bool]
            v[:,not_u_bool] = z[:,not_u_bool]
            f_v = self.integrand.f(v,compute_flags=flags_k,*args,**kwargs)
            y[:,0,0,k] = f_x*(f_v-f_z) # A.18
            y[:,1,0,k] = (f_z-f_v)**2/2 # A.16
            y[:,:,1,k] = f_x[:,None,:] # mu
            y[:,:,2,k] = f_x[:,None,:]**2 # sigma^2+mu^2
        return y
    
    def _spawn(self, level, sampler):
        new_integrand = self.integrand.spawn(level,sampler)
        return SensitivityIndices(
            integrand = new_integrand,
            indices = self.indices)
    
    def bound_fun(self, bound_low, bound_high):
        tau_low,mu_low,f2_low = bound_low[:,0],bound_low[:,1],bound_low[:,2]
        tau_high,mu_high,f2_high = bound_high[:,0],bound_high[:,1],bound_high[:,2]
        sigma2_low1,sigma2_low2 = f2_high-mu_low**2,f2_high-mu_high**2
        comb_bounds_low = clip(minimum.reduce([tau_low/sigma2_low1,tau_low/sigma2_low2]),0,1)
        sigma2_high1,sigma2_high2 = f2_low-mu_low**2,f2_low-mu_high**2
        comb_bounds_high = clip(maximum.reduce([tau_high/sigma2_high1,tau_high/sigma2_high2]),0,1)
        violated = (sigma2_high1<=0)|(sigma2_high2<=0)
        comb_bounds_low[violated],comb_bounds_high[violated] = 0,1
        return comb_bounds_low,comb_bounds_high
    
    def dependency(self, comb_flags):
        return repeat(comb_flags[:,None,:],3,axis=1)

class SobolIndices(SensitivityIndices):
    """ Normalized Sobol' Indices, an alias for SensitivityIndices. """
    pass


from ._true_measure import TrueMeasure
from ..util import TransformError, DimensionError
from ..discrete_distribution import DigitalNetB2
from numpy import *
from scipy.stats import norm


class Uniform(TrueMeasure):
    """
    >>> u = Uniform(DigitalNetB2(2,seed=7),lower_bound=[0,.5],upper_bound=[2,3])
    >>> u.gen_samples(4)
    array([[0.1431124 , 0.69460269],
           [1.62840337, 2.36213894],
           [0.62818597, 2.83084783],
           [1.14326115, 1.16339383]])
    >>> u
    Uniform (TrueMeasure Object)
        lower_bound     [0.  0.5]
        upper_bound     [2 3]
    """
    
    def __init__(self, sampler, lower_bound=0., upper_bound=1.):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform 
            lower_bound (float): a for Uniform(a,b)
            upper_bound (float): b for Uniform(a,b)
        """
        self.parameters = ['lower_bound', 'upper_bound']
        self.domain = array([[0,1]])
        self._parse_sampler(sampler)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        if isscalar(self.lower_bound):
            lower_bound = tile(self.lower_bound,self.d)
        if isscalar(self.upper_bound):
            upper_bound = tile(self.upper_bound,self.d)
        self.a = array(lower_bound)
        self.b = array(upper_bound)
        if len(self.a)!=self.d or len(self.b)!=self.d:
            raise DimensionError('upper bound and lower bound must be of length dimension')
        self._set_constants()
        self.range = hstack((self.a.reshape((self.d,1)),self.b.reshape((self.d,1))))
        super(Uniform,self).__init__()

    def _set_constants(self):
        self.delta = self.b - self.a
        self.inv_delta_prod = 1/self.delta.prod()

    def _transform(self, x):
        return x * self.delta + self.a
    
    def _weight(self, x):
        return tile(self.inv_delta_prod,x.shape[0])
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = Uniform(sampler,lower_bound=self.a,upper_bound=self.b)
        else:
            l = self.a[0]
            u = self.b[0]
            if not (all(self.a==l) and all(self.b==u)):
                raise DimensionError('''
                    In order to spawn a uniform measure
                    the lower bounds must all be the same and 
                    the upper bounds must all be the same''')
            spawn = Uniform(sampler,lower_bound=l,upper_bound=u)
        return spawn
    


from ..integrand._integrand import Integrand
from ..util import DistributionCompatibilityError, ParameterError, MethodImplementationError, _univ_repr
from numpy import *

class StoppingCriterion(object):
    """ Stopping Criterion abstract class. DO NOT INSTANTIATE. """
    
    def __init__(self, allowed_levels, allowed_distribs, allow_vectorized_integrals):
        """
        Args:
            distribution (DiscreteDistribution): a DiscreteDistribution
            allowed_levels (list): which integrand types are supported: 'single', 'fixed-multi', 'adaptive-multi'
            allowed_distribs (list): list of compatible DiscreteDistribution classes
        """
        sname = type(self).__name__
        prefix = 'A concrete implementation of StoppingCriterion must have '
        # integrand check
        if (not hasattr(self, 'integrand')) or \
            (not isinstance(self.integrand,Integrand)):
            raise ParameterError(prefix + 'self.integrand, an Integrand instance')
        # true measure check
        if (not hasattr(self, 'true_measure')) or (self.true_measure!=self.integrand.true_measure):
            raise ParameterError(prefix + 'self.true_measure=self.integrand.true_measure')
        # discrete distribution check
        if (not hasattr(self, 'discrete_distrib')) or (self.discrete_distrib!=self.integrand.discrete_distrib):
            raise ParameterError(prefix + 'self.discrete_distrib=self.integrand.discrete_distrib')
        if not isinstance(self.discrete_distrib,tuple(allowed_distribs)):
            raise DistributionCompatibilityError('%s must have a DiscreteDistribution in %s'%(sname,str(allowed_distribs)))
        # multilevel compatibility check
        if self.integrand.leveltype not in allowed_levels:
            raise ParameterError('Integrand is %s level but %s only supports %s level problems.'%(self.integrand.leveltype,sname,allowed_levels))
        if (not allow_vectorized_integrals) and self.integrand.d_indv!=(1,):
            raise ParameterError('Vectorized integrals (with d_indv>1 outputs per sample) are not supported by this stopping criterion')
        # parameter checks
        if not hasattr(self,'parameters'):
            self.parameters = []
            
    def integrate(self):
        """
        ABSTRACT METHOD to determine the number of samples needed to satisfy the tolerance.

        Return:
            tuple: tuple containing:
                - solution (float): approximation to the integral
                - data (AccumulateData): an AccumulateData object
        """
        raise MethodImplementationError(self, 'integrate')
    
    def set_tolerance(self, *args, **kwargs):
        """ ABSTRACT METHOD to reset the absolute tolerance. """
        raise ParameterError("The %s StoppingCriterion does not yet support resetting tolerances.")

    def _compute_indv_alphas(self, alphas_comb):
        """
        Compute individual uncertainty levels required to achieve combined uncertainty levels. 

        Args:
            alphas_comb (ndarray): desired uncertainty levels on combined solutions. 
        
        Return:
            ndarray: uncertainty levels on individual solutions"""
        alphas_indv = tile(1,self.integrand.d_indv)
        identity_dependency = True
        for k in ndindex(self.integrand.d_comb):
            comb_flags = tile(True,self.integrand.d_comb)
            comb_flags[k] = False
            flags_indv = self.integrand.dependency(comb_flags)
            if self.integrand.d_indv!=self.integrand.d_comb or (flags_indv!=comb_flags).any(): identity_dependency=False
            dependents_k = ~flags_indv
            n_dep_k = dependents_k.sum()
            alpha_k = alphas_comb[k]/n_dep_k
            alpha_k_mat = alpha_k*dependents_k
            alpha_k_mat[alpha_k_mat==0] = 1
            alphas_indv = minimum(alphas_indv,alpha_k_mat)
        return alphas_indv,identity_dependency
    def __repr__(self):
        return _univ_repr(self, "StoppingCriterion", self.parameters)



from numpy import *
from ._integrand import Integrand
from ..true_measure import Uniform
from ..discrete_distribution import DigitalNetB2

class Sin1d(Integrand):
    """
    >>> sin1d = Sin1d(DigitalNetB2(1,seed=7))
    >>> x = sin1d.discrete_distrib.gen_samples(2**10)
    >>> y = sin1d.f(x)
    >>> y.mean()
    -7.3731683...e-08
    >>> sin1d.true_measure
    Uniform (TrueMeasure Object)
        lower_bound     0
        upper_bound     6.283
    """
    def __init__(self, sampler, k=1):
        self.sampler = sampler
        self.k = k
        assert self.sampler.d==1
        self.true_measure = Uniform(self.sampler,lower_bound=0,upper_bound=2*self.k*pi)
        super(Sin1d,self).__init__(dimension_indv=1,dimension_comb=1,parallel=False)
    def g(self, t):
        return sin(t.squeeze())
    def _spawn(self, level, sampler):
        return Sin1d(sampler=sampler,k=self.k)


from ._integrand import Integrand
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Uniform
from numpy import *


class BoxIntegral(Integrand):
    """
    $B_s(x) = \\left(\\sum_{j=1}^d x_j^2 \\right)^{s/2}$

    >>> l1 = BoxIntegral(DigitalNetB2(2,seed=7), s=[7])
    >>> x1 = l1.discrete_distrib.gen_samples(2**10)
    >>> y1 = l1.f(x1)
    >>> y1.shape
    (1024, 1)
    >>> y1.mean(0)
    array([0.75147585])
    >>> l2 = BoxIntegral(DigitalNetB2(5,seed=7), s=[-7,7])
    >>> x2 = l2.discrete_distrib.gen_samples(2**10)
    >>> y2 = l2.f(x2,compute_flags=[1,1])
    >>> y2.shape
    (1024, 2)
    >>> y2.mean(0)
    array([ 9.2238684 , 10.50687484])

    References:

    [1] D.H. Bailey, J.M. Borwein, R.E. Crandall, Box integrals,
    Journal of Computational and Applied Mathematics, Volume 206, Issue 1, 2007, Pages 196-208, ISSN 0377-0427, 
    https://doi.org/10.1016/j.cam.2006.06.010. (https://www.sciencedirect.com/science/article/pii/S0377042706004250) 
    
    [2] https://www.davidhbailey.com/dhbpapers/boxintegrals.pdf
    """

    def __init__(self, sampler, s=array([1,2])):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform
            s (list or ndarray): vectorized s parameter, len(s) is the number of vectorized integrals to evaluate.
        """
        self.parameters = ['s']
        self.s = array([s]) if isscalar(s) else array(s)
        self.sampler = sampler
        self.true_measure = Uniform(self.sampler, lower_bound=0., upper_bound=1.)
        super(BoxIntegral,self).__init__(dimension_indv=len(self.s),dimension_comb=len(self.s),parallel=False) # output dimensions per sample

    def g(self, t, **kwargs):
        compute_flags = kwargs['compute_flags'] if 'compute_flags' in kwargs else ones(self.d_indv,dtype=int)
        n,d = t.shape
        y = zeros((n,)+self.d_indv,dtype=float)
        for j in range(len(self.s)):
            if compute_flags[j] == 1: 
                y[:,j] = (t**2).sum(1)**(self.s[j]/2)
        return y
    
    def _spawn(self, level, sampler):
        return BoxIntegral(sampler=sampler,s=self.s)



from ._stopping_criterion import StoppingCriterion
from ..accumulate_data import MeanVarDataVec
from ..discrete_distribution._discrete_distribution import DiscreteDistribution
from ..discrete_distribution import IIDStdUniform
from ..discrete_distribution._discrete_distribution import IID
from ..true_measure import Gaussian,Uniform
from ..integrand import Keister,BoxIntegral,CustomFun
from ..util import MaxSamplesWarning, ParameterWarning, ParameterError
from numpy import *
from time import time
from scipy.stats import norm
import warnings


class CubMCCLTVec(StoppingCriterion):
    """
    Stopping criterion based on the Central Limit Theorem for vectorized integrands.
    
    >>> k = Keister(IIDStdUniform(seed=7))
    >>> sc = CubMCCLTVec(k,abs_tol=.05)
    >>> solution,data = sc.integrate()
    >>> solution
    array([1.38366791])
    >>> data
    MeanVarDataVec (AccumulateData Object)
        solution        1.384
        comb_bound_low  1.335
        comb_bound_high 1.432
        comb_flags      1
        n_total         2^(10)
        n               2^(10)
        time_integrate  ...
    CubMCCLTVec (StoppingCriterion Object)
        inflate         1.200
        alpha           0.010
        abs_tol         0.050
        rel_tol         0
        n_init          2^(8)
        n_max           2^(30)
    Keister (Integrand Object)
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      2^(-1)
        decomp_type     PCA
    IIDStdUniform (DiscreteDistribution Object)
        d               1
        entropy         7
        spawn_key       ()
    >>> f = BoxIntegral(IIDStdUniform(3,seed=7), s=[-1,1])
    >>> abs_tol = 5e-2
    >>> sc = CubMCCLTVec(f,abs_tol=abs_tol)
    >>> solution,data = sc.integrate()
    >>> solution
    array([1.1853359 , 0.95670595])
    >>> data
    MeanVarDataVec (AccumulateData Object)
        solution        [1.185 0.957]
        comb_bound_low  [1.139 0.918]
        comb_bound_high [1.232 0.995]
        comb_flags      [ True  True]
        n_total         2^(11)
        n               [2048.  512.]
        time_integrate  ...
    CubMCCLTVec (StoppingCriterion Object)
        inflate         1.200
        alpha           0.010
        abs_tol         0.050
        rel_tol         0
        n_init          2^(8)
        n_max           2^(30)
    BoxIntegral (Integrand Object)
        s               [-1  1]
    Uniform (TrueMeasure Object)
        lower_bound     0
        upper_bound     1
    IIDStdUniform (DiscreteDistribution Object)
        d               3
        entropy         7
        spawn_key       ()
    >>> sol3neg1 = -pi/4-1/2*log(2)+log(5+3*sqrt(3))
    >>> sol31 = sqrt(3)/4+1/2*log(2+sqrt(3))-pi/24
    >>> true_value = array([sol3neg1,sol31])
    >>> assert (abs(true_value-solution)<abs_tol).all()
    >>> cf = CustomFun(
    ...     true_measure = Uniform(IIDStdUniform(6,seed=7)),
    ...     g = lambda x,compute_flags=None: (2*arange(1,7)*x).reshape(-1,2,3),
    ...     dimension_indv = (2,3))
    >>> sol,data = CubMCCLTVec(cf,abs_tol=1e-2).integrate()
    >>> data
    MeanVarDataVec (AccumulateData Object)
        solution        [[1.    1.999 2.999]
                        [4.001 4.998 6.001]]
        comb_bound_low  [[0.99  1.989 2.991]
                        [3.991 4.989 5.993]]
        comb_bound_high [[1.01  2.009 3.006]
                        [4.01  5.007 6.008]]
        comb_flags      [[ True  True  True]
                        [ True  True  True]]
        n_total         2^(21)
        n               [[  32768.  131072.  524288.]
                        [ 524288. 1048576. 2097152.]]
        time_integrate  ...
    CubMCCLTVec (StoppingCriterion Object)
        inflate         1.200
        alpha           0.010
        abs_tol         0.010
        rel_tol         0
        n_init          2^(8)
        n_max           2^(30)
    CustomFun (Integrand Object)
    Uniform (TrueMeasure Object)
        lower_bound     0
        upper_bound     1
    IIDStdUniform (DiscreteDistribution Object)
        d               6
        entropy         7
        spawn_key       ()
    """

    def __init__(self, integrand, abs_tol=1e-2, rel_tol=0., n_init=256., n_max=2**30,
        inflate=1.2, alpha=0.01,
        error_fun = lambda sv,abs_tol,rel_tol: maximum(abs_tol,abs(sv)*rel_tol)):
        """
        Args:
            integrand (Integrand): an instance of Integrand
            inflate (float): inflation factor when estimating variance
            alpha (ndarray): significance level for confidence interval
            abs_tol (ndarray): absolute error tolerance
            rel_tol (ndarray): relative error tolerance
            n_max (int): maximum number of samples
            error_fun: function taking in the approximate solution vector, 
                absolute tolerance, and relative tolerance which returns the approximate error. 
                Default indicates integration until either absolute OR relative tolerance is satisfied.
        """
        self.parameters = ['inflate','alpha','abs_tol','rel_tol','n_init','n_max']
        # Input Checks
        if log2(n_init) % 1 != 0:
            warning_s = ' n_init must be a power of 2. Using n_init = 32'
            warnings.warn(warning_s, ParameterWarning)
            n_init = 32
        # Set Attributes
        self.abs_tol = abs_tol
        self.rel_tol = rel_tol
        self.n_init = float(n_init)
        self.n_max = float(n_max)
        self.alpha = alpha
        self.inflate = float(inflate)
        self.error_fun = error_fun
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        self.d_indv = self.integrand.d_indv
        self.d = self.discrete_distrib.d
        super(CubMCCLTVec,self).__init__(allowed_levels=["single"], allowed_distribs=[IID], allow_vectorized_integrals=True)
        self.alphas_indv,identity_dependency = self._compute_indv_alphas(full(self.integrand.d_comb,self.alpha))
        self.z_star = -norm.ppf(self.alphas_indv/2)
         
    def integrate(self):
        """ See abstract method. """
        t_start = time()
        self.datum = empty(self.d_indv,dtype=object)
        for j in ndindex(self.d_indv):
            self.datum[j] = MeanVarDataVec(self.z_star[j],self.inflate)
        self.data = MeanVarDataVec.__new__(MeanVarDataVec)
        self.data.flags_indv = tile(False,self.d_indv)
        self.data.compute_flags = tile(True,self.d_indv)
        self.data.indv_bound_low = tile(-inf,self.d_indv)
        self.data.indv_bound_high = tile(inf,self.d_indv)
        self.data.solution_indv = tile(nan,self.d_indv)
        self.data.solution = nan
        self.data.xfull = empty((0,self.d))
        self.data.yfull = empty((0,)+self.d_indv)
        n_min,n_max = 0,self.n_init
        self.data.n = tile(self.n_init,self.d_indv)
        while True:
            n = int(n_max-n_min)
            xnext = self.discrete_distrib.gen_samples(n)
            ynext = self.integrand.f(xnext,compute_flags=self.data.compute_flags)
            self.data.xfull = vstack((self.data.xfull,xnext))
            self.data.yfull = vstack((self.data.yfull,ynext))
            for j in ndindex(self.d_indv):
                if self.data.flags_indv[j]: continue
                yj = self.data.yfull[(slice(None),)+j]
                self.data.solution_indv[j],self.data.indv_bound_low[j],self.data.indv_bound_high[j] = self.datum[j].update_data(yj)
            self.data.comb_bound_low,self.data.comb_bound_high = self.integrand.bound_fun(self.data.indv_bound_low,self.data.indv_bound_high)
            self.abs_tols,self.rel_tols = full_like(self.data.comb_bound_low,self.abs_tol),full_like(self.data.comb_bound_low,self.rel_tol)
            fidxs = isfinite(self.data.comb_bound_low)&isfinite(self.data.comb_bound_high)
            slow,shigh,abs_tols,rel_tols = self.data.comb_bound_low[fidxs],self.data.comb_bound_high[fidxs],self.abs_tols[fidxs],self.rel_tols[fidxs]
            self.data.solution = tile(nan,self.data.comb_bound_low.shape)
            self.data.solution[fidxs] = 1/2*(slow+shigh+self.error_fun(slow,abs_tols,rel_tols)-self.error_fun(shigh,abs_tols,rel_tols))
            self.data.comb_flags = tile(False,self.data.comb_bound_low.shape)
            self.data.comb_flags[fidxs] = (shigh-slow) <= (self.error_fun(slow,abs_tols,rel_tols)+self.error_fun(shigh,abs_tols,rel_tols))
            self.data.flags_indv = self.integrand.dependency(self.data.comb_flags)
            self.data.compute_flags = ~self.data.flags_indv
            self.data.n_total = self.data.n.max()
            if sum(self.data.compute_flags)==0:
                break # sufficiently estimated
            elif 2*self.data.n_total>self.n_max:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples would exceeds n_max = %d.
                No more samples will be generated.
                Note that error tolerances may not be satisfied. """ \
                % (int(self.data.n_total),int(self.data.n_total),int(self.n_max))
                warnings.warn(warning_s, MaxSamplesWarning)
                break
            else:
                n_min = n_max; n_max = 2*n_min
                self.data.n += self.data.n*(self.data.compute_flags) # double sample size
        self.data.integrand = self.integrand
        self.data.true_measure = self.true_measure
        self.data.discrete_distrib = self.discrete_distrib
        self.data.stopping_crit = self
        self.data.parameters = [
            'solution',
            'comb_bound_low',
            'comb_bound_high',
            'comb_flags',
            'n_total',
            'n',
            'time_integrate']
        self.data.datum = self.datum
        self.data.time_integrate = time()-t_start
        return self.data.solution,self.data
    
    def set_tolerance(self, abs_tol=None, rel_tol=None):
        """
        See abstract method. 
        
        Args:
            abs_tol (float): absolute tolerance. Reset if supplied, ignored if not. 
            rel_tol (float): relative tolerance. Reset if supplied, ignored if not. 
        """
        if abs_tol != None: self.abs_tol = abs_tol
        if rel_tol != None: self.rel_tol = rel_tol
        
        


    


from ._integrand import Integrand
from ..true_measure import BrownianMotion
from ..discrete_distribution import DigitalNetB2
from ..util import ParameterError
from numpy import *
from scipy.stats import norm 


class EuropeanOption(Integrand):
    """
    European financial option. 

    >>> eo = EuropeanOption(DigitalNetB2(4,seed=7),call_put='put')
    >>> eo
    EuropeanOption (Integrand Object)
        volatility      2^(-1)
        call_put        put
        start_price     30
        strike_price    35
        interest_rate   0
    >>> x = eo.discrete_distrib.gen_samples(2**12)
    >>> y = eo.f(x)
    >>> y.mean()
    9.212...
    >>> eo = EuropeanOption(BrownianMotion(DigitalNetB2(4,seed=7),drift=1),call_put='put')
    >>> x = eo.discrete_distrib.gen_samples(2**12)
    >>> y = eo.f(x)
    >>> y.mean()
    9.181...
    >>> eo.get_exact_value()
    9.211452976234058
    """
                          
    def __init__(self, sampler, volatility=0.5, start_price=30, strike_price=35,
        interest_rate=0, t_final=1, call_put='call'):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform
            volatility (float): sigma, the volatility of the asset
            start_price (float): S(0), the asset value at t=0
            strike_price (float): strike_price, the call/put offer
            interest_rate (float): r, the annual interest rate
            t_final (float): exercise time
            call_put (str): 'call' or 'put' option
        """
        self.parameters = ['volatility', 'call_put', 'start_price', 'strike_price', 'interest_rate']
        self.t_final = t_final
        self.sampler = sampler
        self.true_measure = BrownianMotion(self.sampler,t_final=self.t_final)
        self.volatility = float(volatility)
        self.start_price = float(start_price)
        self.strike_price = float(strike_price)
        self.interest_rate = float(interest_rate)
        self.call_put = call_put.lower()
        if self.call_put not in ['call','put']:
            raise ParameterError("call_put must be either 'call' or 'put'")
        super(EuropeanOption,self).__init__(dimension_indv=1,dimension_comb=1,parallel=False)  

    def g(self, t):
        """ See abstract method. """
        self.s = self.start_price * exp(
            (self.interest_rate - self.volatility ** 2 / 2) *
            self.true_measure.time_vec + self.volatility * t)
        for xx,yy in zip(*where(self.s<0)): # if stock becomes <=0, 0 out rest of path
            self.s[xx,yy:] = 0
        if self.call_put == 'call':
            y_raw = maximum(self.s[:,-1] - self.strike_price, 0)
        else: # put
            y_raw = maximum(self.strike_price - self.s[:,-1], 0)
        y_adj = y_raw * exp(-self.interest_rate * self.t_final)
        return y_adj
    
    def get_exact_value(self):
        """
        Get the fair price of a European call/put option.
        
        Return:
            float: fair price
        """
        denom = self.volatility * sqrt(self.t_final)
        decay = self.strike_price * exp(-self.interest_rate * self.t_final)
        if self.call_put == 'call':
            term1 = log(self.start_price / self.strike_price) + \
                    (self.interest_rate + self.volatility**2/2) * self.t_final
            term2 = log(self.start_price / self.strike_price) + \
                    (self.interest_rate - self.volatility**2/2) * self.t_final
            fp = self.start_price * norm.cdf(term1/denom) - decay * norm.cdf(term2/denom)
        elif self.call_put == 'put':
            term1 = log(self.strike_price / self.start_price) - \
                    (self.interest_rate - self.volatility**2/2) * self.t_final
            term2 = log(self.strike_price / self.start_price) - \
                    (self.interest_rate + self.volatility**2/2) * self.t_final
            fp = decay * norm.cdf(term1/denom) - self.start_price * norm.cdf(term2/denom)
        return fp
    
    def _spawn(self, level, sampler):
        return EuropeanOption(
            sampler=sampler,
            volatility=self.volatility,
            start_price=self.start_price,
            strike_price=self.strike_price,
            interest_rate=self.interest_rate,
            t_final=self.t_final,
            call_put=self.call_put)



from ._accumulate_data import AccumulateData
from ..util import ExactGPyTorchRegressionModel
from numpy import *
from scipy.stats import norm
import warnings
import gpytorch 
import torch

def _get_phi(gp, x):
    yhat,yhatstd = gp.predict(x)
    with errstate(all='ignore'): z = yhat/yhatstd
    return norm.cdf(z)
def _error_udens_from_phi(phi):
    return 2*minimum(1-phi,phi)
def _error_udens(gp, x):
    return _error_udens_from_phi(_get_phi(gp,x))

class PFGPCIData(AccumulateData):
    """ Update and store data for the PFGPCI StoppingCriterion. """

    def __init__(self, stopping_crit, integrand, true_measure, discrete_distrib, dnb2, n_approx, alpha, refit, 
        gpytorch_prior_mean, gpytorch_prior_cov, gpytorch_likelihood, gpytorch_marginal_log_likelihood_func,
        torch_optimizer_func,gpytorch_train_iter,gpytorch_use_gpu,verbose,approx_true_solution):
        self.parameters = ['solution','error_bound','bound_low','bound_high','n_total','time_integrate']
        self.stopping_crit = stopping_crit
        self.integrand = integrand
        self.true_measure = true_measure
        self.discrete_distrib = discrete_distrib
        self.d = self.discrete_distrib.d
        self.dnb2 = dnb2
        self.alpha = alpha
        self.refit = refit
        self.qmc_pts = self.dnb2.gen_samples(n_approx)
        self.gpytorch_prior_mean = gpytorch_prior_mean
        self.gpytorch_prior_cov = gpytorch_prior_cov
        self.gpytorch_likelihood = gpytorch_likelihood
        self.gpytorch_marginal_log_likelihood_func = gpytorch_marginal_log_likelihood_func
        self.torch_optimizer_func = torch_optimizer_func
        self.gpytorch_train_iter = gpytorch_train_iter
        self.gpytorch_use_gpu = gpytorch_use_gpu
        self.verbose = verbose
        self.approx_true_solution = approx_true_solution
        self.n_batch = []
        self.emr = []
        self.error_bounds = []
        self.ci_low = []
        self.ci_high = []
        self.solutions = []
        self.x = empty((0,self.d),dtype=float)
        self.y = array([],dtype=float)
        self.gpyt_model = ExactGPyTorchRegressionModel(
            x_t = self.x, y_t = self.y,
            prior_mean = self.gpytorch_prior_mean,
            prior_cov = self.gpytorch_prior_cov,
            likelihood = self.gpytorch_likelihood,
            use_gpu = self.gpytorch_use_gpu)   
        self.saved_gps = [self.gpyt_model.state_dict()]
        super(PFGPCIData,self).__init__()

    def update_data(self, batch_count, xdraw, ydrawtf):
        self.n_batch.append(len(xdraw))
        self.x,self.y = vstack([self.x,xdraw]),hstack([self.y,ydrawtf])
        if batch_count==0 or self.refit:
            self.gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x, y_t = self.y,
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            self.gpyt_model.fit(
                optimizer = self.torch_optimizer_func(self.gpyt_model),
                mll = self.gpytorch_marginal_log_likelihood_func(self.gpyt_model.likelihood,self.gpyt_model),
                training_iter = self.gpytorch_train_iter,
                verbose = self.verbose)
        else:
            self.gpyt_model = self.gpyt_model.add_data(xdraw,ydrawtf)
            torch.cuda.empty_cache()          
        self.saved_gps.append(self.gpyt_model.state_dict())
        phi = _get_phi(self.gpyt_model,self.qmc_pts)
        self.solutions.append(mean(phi>=.5))
        self.emr.append(mean(minimum(phi,1-phi)))
        gamma = self.emr[-1]/self.alpha
        self.ci_low.append(max(self.solutions[-1]-gamma,0))
        self.ci_high.append(min(self.solutions[-1]+gamma,1))
        self.error_bounds.append(max(self.solutions[-1]-self.ci_low[-1],self.ci_high[-1]-self.solutions[-1]))

    def get_results_dict(self):
        df = {'n_sum':self.n_sum, 'n_batch':self.n_batch, 'error_bounds':self.error_bounds, 'ci_low':self.ci_low, 'ci_high':array(self.ci_high), 'solutions':array(self.solutions)}
        if self.approx_true_solution: df['solutions_ref'],df['error_ref'],df['in_ci'] = self.solutions_ref,self.error_ref,self.in_ci
        return df 
    
    def plot(self, trace_only=False, **kwargs):
        from matplotlib import pyplot
        if self.d==1 and not trace_only and self.saved_gps!=[]: fig,gs = self.plot_1d(**kwargs)
        elif self.d==2 and not trace_only and self.saved_gps!=[]: fig,gs = self.plot_2d(**kwargs)
        else:
            from matplotlib import pyplot,gridspec
            fig = pyplot.figure(constrained_layout=False,figsize=(8,4))
            gs = gridspec.GridSpec(1,2,figure=fig)
        axtrace = fig.add_subplot(gs[-1,:])
        if self.approx_true_solution: axtrace.plot(.5+arange(len(self.n_batch)),self.solutions_ref,color='c',label=r'$P(g)$')
        axtrace.plot(.5+arange(len(self.n_batch)),self.solutions,'-o',color='k',label=r'$\hat{P}_n^\mathrm{QMC}$')
        axtrace.fill_between(.5+arange(len(self.n_batch)),self.ci_low,self.ci_high,color='k',alpha=.15)
        axtrace.plot(.5+arange(len(self.n_batch)),self.error_bounds,'-o',color='m',label=r'$\hat{\gamma}_n^{\text{QMC}}$')
        if self.approx_true_solution: axtrace.plot(.5+arange(len(self.n_batch)),self.error_ref,'-o',color='b',label=r'$\vert \hat{P}_n^{\text{QMC}} - P(g) \vert$')
        axtrace.set_xlim([0,len(self.n_batch)])
        axtrace.set_xticks([])
        axtrace.set_yscale('log',base=10)
        #axtrace.set_ylim([0,1]); axtrace.set_yticks([0,1])
        for spine in ['top','right','bottom']: axtrace.spines[spine].set_visible(False)
        axtrace.legend(loc='lower left',bbox_to_anchor=(0.05, -.2, .9, .102),mode='expand',ncol=4,prop={'size':18},frameon=False)
        return fig
    def plot_1d(self, meshticks=1025, ci_percentage=.95, **kwargs):
        from matplotlib import pyplot,gridspec
        beta = norm.ppf(mean([ci_percentage,1]))
        n_batches = len(self.n_batch)
        fig = pyplot.figure(constrained_layout=False,figsize=(4*n_batches,4*3))
        gs = gridspec.GridSpec(3,n_batches,figure=fig)
        xticks = linspace(0,1,meshticks)[1:-1]
        if self.approx_true_solution: 
            yticks = self.integrand.f(xticks[:,None]).squeeze()
            ytickstf = self.stopping_crit._affine_tf(yticks)
        for j in range(n_batches):
            i0,i1 = 0 if j==0 else self.n_sum[j-1],self.n_sum[j]
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i0], y_t = self.y[:i0],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j])
            ax0j = fig.add_subplot(gs[0,j])
            mr_udens = _error_udens(gpyt_model,xticks[:,None])
            ax0j.fill_between(xticks,mr_udens,color='k',alpha=.5)
            ax0j.scatter(self.x[i0:i1].squeeze(),zeros(i1-i0)+.015,color='r',s=25)
            ax0j.set_ylim([0,1])
            ax0j.set_yticks([0,1])
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i1], y_t = self.y[:i1],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j+1])
            ax1j = fig.add_subplot(gs[1,j],sharey=None if j==0 else ax1j)
            if self.approx_true_solution: ax1j.plot(xticks,ytickstf,color='c',label=r'f(x)',linewidth=2)
            gp_mean,gp_std = gpyt_model.predict(xticks[:,None])
            ax1j.plot(xticks,gp_mean,color='k',label='gp')
            ax1j.fill_between(xticks,gp_mean-beta*gp_std,gp_mean+beta*gp_std,color='k',alpha=.25)
            ax1j.axhline(y=0,color='lightgreen')
            ax1j.scatter(self.x[:i0].squeeze(),self.y[:i0],color='b',label='old pts')
            ax1j.scatter(self.x[i0:i1].squeeze(),self.y[i0:i1],color='r',label='new pts')
            ax1j.set_xlabel(r'$u$')
            if j==0:
                #ax1j.set_ylabel(r'$y$')
                ax0j.set_ylabel(r'$2\mathrm{ERR}_n(u)$')
        for ax in fig.axes:
            ax.set_xlim([0,1])
            ax.set_xticks([0,1])
        return fig,gs
    def plot_2d(self, meshticks=257, clevels=32, **kwargs):
        from matplotlib import pyplot,gridspec,cm
        n_batches = len(self.n_batch)
        fig = pyplot.figure(constrained_layout=False,figsize=(4*n_batches,4*5))
        gs = gridspec.GridSpec(5 if self.approx_true_solution else 4,n_batches,figure=fig)
        x0mesh,x1mesh = meshgrid(linspace(0,1,meshticks)[1:-1],linspace(0,1,meshticks)[1:-1])
        xquery = vstack([x0mesh.flatten(),x1mesh.flatten()]).T
        if self.approx_true_solution:
            yquery = self.integrand.f(xquery).squeeze()
            ymeshtf = self.stopping_crit._affine_tf(yquery).reshape(x0mesh.shape)
        for j in range(n_batches):
            row_idx = 0
            i0,i1 = 0 if j==0 else self.n_sum[j-1],self.n_sum[j]
            if self.approx_true_solution:
                ax0j = fig.add_subplot(gs[row_idx,j])
                ax0j.contourf(x0mesh,x1mesh,ymeshtf,cmap=cm.Greys,vmin=ymeshtf.min(),vmax=ymeshtf.max(),levels=clevels)
                ax0j.contour(x0mesh,x1mesh,ymeshtf,colors='lightgreen',levels=[0],linewidths=5)
                ax0j.set_title(r'$g(\boldsymbol{u})$')
                row_idx += 1
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i0], y_t = self.y[:i0],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j])
            ax1j = fig.add_subplot(gs[row_idx,j])
            udens_mr = _error_udens(gpyt_model,xquery).reshape(x0mesh.shape)
            ax1j.contourf(x0mesh,x1mesh,udens_mr,cmap=cm.Greys,levels=clevels,vmin=0,vmax=1)
            ax1j.scatter(self.x[i0:i1,0],self.x[i0:i1,1],color='r')
            ax1j.set_title(r'$2\mathrm{ERR}_n(\boldsymbol{u})$')
            row_idx += 1
            gpyt_model = ExactGPyTorchRegressionModel(
                x_t = self.x[:i0], y_t = self.y[:i0],
                prior_mean = self.gpytorch_prior_mean,
                prior_cov = self.gpytorch_prior_cov,
                likelihood = self.gpytorch_likelihood,
                use_gpu = self.gpytorch_use_gpu)
            gpyt_model.load_state_dict(self.saved_gps[j+1])
            gp_mean,gp_std = gpyt_model.predict(xquery)
            gp_mean_mesh,gp_std_mesh = gp_mean.reshape(x0mesh.shape),gp_std.reshape(x0mesh.shape)
            ax2j = fig.add_subplot(gs[row_idx,j])
            ax2j.contourf(x0mesh,x1mesh,gp_mean_mesh,cmap=cm.Greys,vmin=gp_mean.min(),vmax=gp_mean.max(),levels=clevels)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                ax2j.contour(x0mesh,x1mesh,gp_mean_mesh,colors='lightgreen',levels=[0],linewidths=5)
            ax2j.set_title(r'$m_n(\boldsymbol{u})$')
            row_idx += 1
            ax3j = fig.add_subplot(gs[row_idx,j])            
            ax3j.contourf(x0mesh,x1mesh,gp_std_mesh,cmap=cm.Greys,levels=clevels)
            ax3j.set_title(r'$\sigma_n(\boldsymbol{u})$')
            for ax in [ax2j,ax3j]:
                ax.scatter(self.x[:i0,0],self.x[:i0,1],color='b')
                ax.scatter(self.x[i0:i1,0],self.x[i0:i1,1],color='r')
        for ax in fig.axes:
            ax.set_xlim([0,1])
            ax.set_ylim([0,1])
            ax.set_xlabel(r'$u_1$')
            ax.set_aspect(1)
            ax.set_xticks([0,1])
            ax.set_yticks([0,1])
            ax.set_ylabel(r'$u_2$')
        return fig,gs



from .._discrete_distribution import LD
from ...util import ParameterError, ParameterWarning
import qmctoolscl
from numpy import *
from numpy.lib.npyio import DataSource
from os.path import dirname, abspath, isfile
import warnings


class Lattice(LD):
    """
    Quasi-Random Lattice nets in base 2.
    
    >>> l = Lattice(2,seed=7)
    >>> l.gen_samples(4)
    array([[0.04386058, 0.58727432],
           [0.54386058, 0.08727432],
           [0.29386058, 0.33727432],
           [0.79386058, 0.83727432]])
    >>> l.gen_samples(1)
    array([[0.04386058, 0.58727432]])
    >>> l
    Lattice (DiscreteDistribution Object)
        d               2^(1)
        dvec            [0 1]
        randomize       SHIFT
        order           NATURAL
        gen_vec         [     1 182667]
        entropy         7
        spawn_key       ()
    >>> Lattice(dimension=2,randomize=False,order='natural').gen_samples(4, warn=False)
    array([[0.  , 0.  ],
           [0.5 , 0.5 ],
           [0.25, 0.75],
           [0.75, 0.25]])
    >>> Lattice(dimension=2,randomize=False,order='linear').gen_samples(4, warn=False)
    array([[0.  , 0.  ],
           [0.25, 0.75],
           [0.5 , 0.5 ],
           [0.75, 0.25]])
    >>> Lattice(dimension=2,randomize=False,order='gray').gen_samples(4, warn=False)
    array([[0.  , 0.  ],
           [0.5 , 0.5 ],
           [0.75, 0.25],
           [0.25, 0.75]])
    >>> l = Lattice(2,generating_vector=25,seed=55)
    >>> l.gen_samples(4)
    array([[0.84489224, 0.30534549],
           [0.34489224, 0.80534549],
           [0.09489224, 0.05534549],
           [0.59489224, 0.55534549]])
    >>> l
    Lattice (DiscreteDistribution Object)
        d               2^(1)
        dvec            [0 1]
        randomize       SHIFT
        order           NATURAL
        gen_vec         [       1 11961679]
        entropy         55
        spawn_key       ()
    >>> Lattice(dimension=4,randomize=False,seed=353,generating_vector=26).gen_samples(8,warn=False)
    array([[0.   , 0.   , 0.   , 0.   ],
           [0.5  , 0.5  , 0.5  , 0.5  ],
           [0.25 , 0.25 , 0.75 , 0.75 ],
           [0.75 , 0.75 , 0.25 , 0.25 ],
           [0.125, 0.125, 0.875, 0.875],
           [0.625, 0.625, 0.375, 0.375],
           [0.375, 0.375, 0.625, 0.625],
           [0.875, 0.875, 0.125, 0.125]])
    >>> Lattice(dimension=3,randomize=False,generating_vector="LDData/main/lattice/mps.exod2_base2_m20_CKN.txt").gen_samples(8,warn=False)
    array([[0.   , 0.   , 0.   ],
           [0.5  , 0.5  , 0.5  ],
           [0.25 , 0.75 , 0.75 ],
           [0.75 , 0.25 , 0.25 ],
           [0.125, 0.375, 0.375],
           [0.625, 0.875, 0.875],
           [0.375, 0.125, 0.125],
           [0.875, 0.625, 0.625]])
    >>> Lattice(3,seed=7,replications=2).gen_samples(4)
    array([[[0.04386058, 0.58727432, 0.3691824 ],
            [0.54386058, 0.08727432, 0.8691824 ],
            [0.29386058, 0.33727432, 0.1191824 ],
            [0.79386058, 0.83727432, 0.6191824 ]],
    <BLANKLINE>
           [[0.65212985, 0.69669968, 0.10605352],
            [0.15212985, 0.19669968, 0.60605352],
            [0.90212985, 0.44669968, 0.85605352],
            [0.40212985, 0.94669968, 0.35605352]]])
    >>> Lattice(3,seed=7,generating_vector=25,replications=2).gen_samples(4)
    array([[[0.3691824 , 0.65212985, 0.69669968],
            [0.8691824 , 0.15212985, 0.19669968],
            [0.6191824 , 0.90212985, 0.44669968],
            [0.1191824 , 0.40212985, 0.94669968]],
    <BLANKLINE>
           [[0.10605352, 0.63025643, 0.13630282],
            [0.60605352, 0.13025643, 0.63630282],
            [0.35605352, 0.38025643, 0.38630282],
            [0.85605352, 0.88025643, 0.88630282]]])
    
    References:

        [1] Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang,
        Lluis Antoni Jimenez Rugama, Da Li, Jagadeeswaran Rathinavel,
        Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou,
        GAIL: Guaranteed Automatic Integration Library (Version 2.3)
        [MATLAB Software], 2019. Available from http://gailgithub.github.io/GAIL_Dev/

        [2] F.Y. Kuo & D. Nuyens.
        Application of quasi-Monte Carlo methods to elliptic PDEs with random diffusion coefficients
        - a survey of analysis and implementation, Foundations of Computational Mathematics,
        16(6):1631-1696, 2016.
        springer link: https://link.springer.com/article/10.1007/s10208-016-9329-5
        arxiv link: https://arxiv.org/abs/1606.06613

        [3] D. Nuyens, `The Magic Point Shop of QMC point generators and generating
        vectors.` MATLAB and Python software, 2018. Available from
        https://people.cs.kuleuven.be/~dirk.nuyens/

        [4] Constructing embedded lattice rules for multivariate integration
        R Cools, FY Kuo, D Nuyens -  SIAM J. Sci. Comput., 28(6), 2162-2188.

        [5] L'Ecuyer, Pierre & Munger, David. (2015).
        LatticeBuilder: A General Software Tool for Constructing Rank-1 Lattice Rules.
        ACM Transactions on Mathematical Software. 42. 10.1145/2754929.
    """

    def __init__(self, dimension=1, randomize='shift', order='natural', seed=None,
        generating_vector="kuo.lattice-33002-1024-1048576.9125.txt", d_max=None, m_max=None, replications=1):
        r"""
        Args:
            dimension (int or :class:`numpy.ndarray`): dimension of the generator.

                - If an int is passed in, use sequence dimensions [0,...,dimension-1].
                - If an ndarray is passed in, use these dimension indices in the sequence.
            
            randomize (bool): If True, apply shift to generated samples.
                Note: Non-randomized lattice sequence includes the origin.
            order (str): "linear", "natural", or "gray" ordering.
            seed (None or int or :class:`numpy.random.SeedSeq`): seed the random number generator for reproducibility
            generating_vector (:class:`numpy.ndarray`, str, or int): Specify the generating matrix. There are a number of optional input types. 
                
                - An ndarray of integers.
                - A string `generating_vector` with either a relative path from `LDData repository <https://github.com/QMCSoftware/LDData>`__ (e.g., "LDData/main/lattice/mps.exod2_base2_m20_CKN.txt")  or a NumPy file with format "name.d_max.m_max.npy" (e.g., "lattice_vec.3600.20.npy").
                - An odd integer :math:`1 < M < 27` which creates a random generating vector :math:`[1,v_1,v_2,...,v_{\\texttt{d\_max}}]` where :math:`v_i` is a random integer in :math:`{3,5,...,2*M-1}` supporting up to :math:`2^M` points.
            
            d_max (int): maximum dimension
            m_max (int): :math:`2^{\\texttt{m\_max}}` is the max number of supported samples
            replications (int): number of IID randomizations of a pointset

        Note:
            `d_max` and `m_max` are required if `generating_vector` is an ndarray.
            If `generating_vector` is a string (path), `d_max` and `m_max` can be taken from the file name if None.
        """
        self.parameters = ['dvec','randomize','order','gen_vec']
        self.mimics = 'StdUniform'
        self.low_discrepancy = True
        self.replications_gv = 1
        # ordering
        self.order = order.upper()
        if self.order=="MPS": raise ParameterError("lattice in MPS order is no longer supported")
        assert self.order in ['LINEAR','NATURAL','GRAY']
        # generating vector
        if isinstance(generating_vector,ndarray):
            self.gen_vec_og = generating_vector
            if d_max is None or m_max is None:
                raise ParameterError("d_max and m_max must be supplied when generating_vector is a ndarray")
            self.d_max = d_max
            self.m_max = m_max
            if self.gen_vec_og.ndim==1:
                self.gen_vec_og = self.gen_vec_og[None,:]
            else:
                assert self.gen_vec_og.ndim==2, "ndarray generating vector must be 1 or two dimensional"
                self.replications_gv = len(self.gen_vec_og)
        elif isinstance(generating_vector,int):
            self.m_max = min(26,max(2,generating_vector))
            assert isinstance(dimension,int), "random generating vector requires int dimension"
            self.d_max = dimension
        elif isinstance(generating_vector,str):
            parts = generating_vector.split('.')
            root = dirname(abspath(__file__))+'/generating_vectors/'
            repos = DataSource()
            if generating_vector[-4:]==".txt":
                if repos.exists(root+generating_vector):
                    datafile = repos.open(root+generating_vector)
                elif repos.exists(generating_vector):
                    datafile = repos.open(generating_vector)
                elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/lattice/"+generating_vector[7:]):
                    datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/lattice/"+generating_vector[7:])
                elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/"+generating_vector):
                    datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/"+generating_vector)
                else:
                    raise ParameterError("LDData path %s not found"%generating_vector)
                contents = [int(line.rstrip('\n').strip().split("#",1)[0]) for line in datafile.readlines() if line[0]!="#"]
                datafile.close()
                self.d_max = contents[0]
                n_max = contents[1]; assert log2(n_max)%1==0; self.m_max = int(log2(n_max))
                self.gen_vec_og = array(contents[2:],dtype=uint64)[None,:]
            elif isfile(root+generating_vector):
                self.gen_vec_og = load(root+generating_vector).astype(uint64)[None,:]
                self.d_max = int(parts[-3])
                self.m_max = int(parts[-2])
            elif isfile(generating_vector):
                self.gen_vec_og = load(generating_vector).astype(uint64)[None,:]
                self.d_max = int(parts[-3])
                self.m_max = int(parts[-2])
            else:
                raise ParameterError("generating_vector '%s' not found."%generating_vector)
        else:
            raise ParameterError("invalid input for generating_matrices, see documentation and doctests.")
        super(Lattice,self).__init__(dimension,seed)
        # randomization
        if isinstance(generating_vector,int):
            self.replications_gv = replications
            self.gen_vec_og = hstack([ones((replications,1),dtype=uint64),2*self.rng.integers(1,2**(self.m_max-1),size=(replications,dimension-1),dtype=uint64)+1])
        self.gen_vec = self.gen_vec_og[:,self.dvec].copy()
        assert self.gen_vec.ndim==2, "gen_vec must have 2 dimensions"
        self.randomize = str(randomize).upper()
        if self.randomize=="TRUE": self.randomize = "SHIFT"
        if self.randomize=="NONE": self.randomize = "FALSE"
        if self.randomize=="NO": self.randomize = "FALSE"
        assert self.randomize in ["SHIFT","FALSE"]
        if self.randomize=="SHIFT":
            self.shift = self.rng.uniform(size=(replications,self.d))
        if self.replications_gv>1: assert replications==self.replications_gv, "if replications_gv>1 require replications = replications_gv"
        self.replications = replications
        

    def gen_samples(self, n=None, n_min=0, n_max=8, warn=True, return_unrandomized=False, qmctoolscl_gen_kwargs={"backend":"c"}, qmctoolscl_rand_kwargs={"backend":"c"}):
        """
        Generate lattice samples

        Args:
            n (int): if n is supplied, generate from n_min=0 to n_max=n samples.
                Otherwise use the n_min and n_max explicitly supplied as the following 2 arguments
            n_min (int): Starting index of sequence.
            n_max (int): Final index of sequence.
            return_unrandomized (bool): return samples without randomization as 2nd return value.
                Will not be returned if randomize=False.
            qmctoolscl_gen_kwargs,qmctoolscl_rand_kwargs (dict): keyword arguments for QMCToolsCL to use OpenCL when generating points or performing randomizations. Defaults to C backend. See https://qmcsoftware.github.io/QMCToolsCL/

        Returns:
            ndarray: replications x (n_max-n_min) x d (dimension) array of samples

        Note:
            Lattice generates in blocks from 2**m to 2**(m+1) so generating
            n_min=3 to n_max=9 requires necessarily produces samples from n_min=2 to n_max=16
            and automatically subsets. May be inefficient for non-powers-of-2 samples sizes.
        """
        if n:
            n_min = 0
            n_max = n            
        if n_min == 0 and self.randomize=="FALSE" and warn:
            warnings.warn("Non-randomized lattice sequence includes the origin",ParameterWarning)
        if n_max > 2**self.m_max:
            raise ParameterError('Lattice generating vector supports up to %d samples.'%(2**self.m_max))
        r_x = uint64(self.replications_gv) 
        n = uint64(n_max-n_min)
        d = uint64(self.d) 
        n_start = uint64(n_min)
        x = empty((r_x,n,d),dtype=float64)
        if self.order=="LINEAR":
            assert r_x==1, "lattice linear currently requires there be only 1 generating matrix"
            x = self._gail_linear(n_min,n_max)[None,:,:]
        elif self.order=="NATURAL":
            assert (n_min==0 or log2(n_min)%1==0) and (n_max==0 or log2(n_max)%1==0), "lattice in natural order requires n_min and n_max be 0 or powers of 2"
            _ = qmctoolscl.lat_gen_natural(r_x,n,d,n_start,self.gen_vec,x,**qmctoolscl_gen_kwargs)
        elif self.order=="GRAY":
            _ = qmctoolscl.lat_gen_gray(r_x,n,d,n_start,self.gen_vec,x,**qmctoolscl_gen_kwargs) 
        else: 
            assert False, "invalid lattice order"
        if r_x==1: x = x[0]
        r = uint64(self.replications)
        if self.randomize=="FALSE":
            assert return_unrandomized is False, "cannot return_unrandomized when randomize='FALSE'"
            return x
        elif self.randomize=="SHIFT":
            xr = empty((r,n,d),dtype=float64)
            qmctoolscl.lat_shift_mod_1(r,n,d,r_x,x,self.shift,xr,**qmctoolscl_rand_kwargs)
            if r==1: xr=xr[0]
            return (xr,x) if return_unrandomized else xr
        else:
            raise ParameterError("incorrect randomize parsing in lattice gen_samples")

    def _gen_block_linear(self, m_next, first=True):
        n = int(2**m_next)
        if first:
            y = arange(0, 1, 1 / n).reshape((n, 1))
        else:
            y = arange(1 / n, 1, 2 / n).reshape((n, 1))
        x = outer(y, self.gen_vec) % 1
        return x

    def calculate_y(self, m_low, m_high, y):
        for m in range(m_low, m_high):
            n = 2 ** m
            y_next = arange(1 / n, 1, 2 / n).reshape((int(n / 2), 1))
            temp = zeros((n, 1))
            temp[0::2] = y
            temp[1::2] = y_next
            y = temp
        return y

    def _gail_linear(self, n_min, n_max):
        """ Gail lattice generator in linear order. """
        m_low = int(floor(log2(n_min))) + 1 if n_min > 0 else 0
        m_high = int(ceil(log2(n_max)))
        if n_min == 0:
            return self._gen_block_linear(m_high, first=True)
        else:
            n = 2 ** (m_low)
            y = arange(1 / n, 1, 2 / n).reshape((int(n / 2), 1))
            y = self.calculate_y(m_low, m_high, y)
            x = outer(y, self.gen_vec) % 1
            return x
        
    def pdf(self, x):
        """ pdf of a standard uniform """
        return ones(x.shape[:-1], dtype=float)

    def _spawn(self, child_seed, dimension):
        return Lattice(
                dimension=dimension,
                randomize=self.randomize,
                order=self.order,
                seed=child_seed,
                generating_vector=self.gen_vec_og,
                d_max=self.d_max,
                m_max=self.m_max,
                replications=self.replications)


from ._accumulate_data import AccumulateData
from numpy import *
from numpy.linalg import lstsq


class MLMCData(AccumulateData):
    """
    Accumulated data for IIDDistribution calculations,
    and store multi-level mean, variance, and cost values.
    See the stopping criterion that utilize this object for references.
    """

    def __init__(self, stopping_crit, integrand, true_measure, discrete_distrib, levels_init, n_init, 
                alpha0, beta0, gamma0):
        """
        Initialize data instance

        Args:
            stopping_crit (StoppingCriterion): a StoppingCriterion instance
            integrand (Integrand): an Integrand instance
            true_measure (TrueMeasure): A TrueMeasure instance
            discrete_distrib (DiscreteDistribution): a DiscreteDistribution instance
            levels_init (int): initial number of levels
            n_init (int): initial number of samples per level
            alpha0 (float): weak error is O(2^{-alpha0*level})
            beta0 (float): variance is O(2^{-beta0*level})
            gamma0 (float): sample cost is O(2^{gamma0*level})
        """
        self.parameters = ['solution','n_total','levels','n_level','mean_level','var_level', 
            'cost_per_sample','alpha','beta','gamma']
        self.stopping_crit = stopping_crit
        self.integrand = integrand
        self.true_measure = true_measure
        self.discrete_distrib = discrete_distrib
        # Set Attributes
        self.levels = int(levels_init)
        self.n_level = zeros(self.levels+1)
        self.sum_level = zeros((2,self.levels+1))
        self.cost_level = zeros(self.levels+1)
        self.diff_n_level = tile(n_init,self.levels+1)
        self.alpha0 = alpha0
        self.beta0 = beta0
        self.gamma0 = gamma0
        self.alpha = maximum(0,self.alpha0)
        self.beta = maximum(0,self.beta0)
        self.gamma = maximum(0,self.gamma0)
        self.solution = None
        self.n_total = 0
        self.time_integrate = 0
        self.level_integrands = []
        super(MLMCData,self).__init__()

    def update_data(self):
        """ See abstract method. """
        # update sample sums
        for l in range(self.levels+1):
            if l==len(self.level_integrands):
                # haven't spawned this level's integrand yet
                self.level_integrands += self.integrand.spawn(levels=int(l))
            integrand_l = self.level_integrands[l]
            if self.diff_n_level[l] > 0:
                # evaluate integral at sampling points samples
                samples = integrand_l.discrete_distrib.gen_samples(n=self.diff_n_level[l])
                integrand_l.f(samples).squeeze()
                self.n_level[l] = self.n_level[l] + self.diff_n_level[l]
                self.sum_level[0,l] = self.sum_level[0,l] + integrand_l.sums[0]
                self.sum_level[1,l] = self.sum_level[1,l] + integrand_l.sums[1]
                self.cost_level[l] = self.cost_level[l] + integrand_l.cost
        # compute absolute average, variance and cost
        self.mean_level = absolute(self.sum_level[0,:self.levels+1]/self.n_level[:self.levels+1])
        self.var_level = maximum(0,self.sum_level[1,:self.levels+1]/self.n_level[:self.levels+1] - self.mean_level**2)
        self.cost_per_sample = self.cost_level[:self.levels+1]/self.n_level[:self.levels+1]
        # fix to cope with possible zero values for self.mean_level and self.var_level
        # (can happen in some applications when there are few samples)
        for l in range(2,self.levels+1):
            self.mean_level[l] = maximum(self.mean_level[l], .5*self.mean_level[l-1]/2**self.alpha)
            self.var_level[l] = maximum(self.var_level[l], .5*self.var_level[l-1]/2**self.beta)
        # use linear regression to estimate alpha, beta, gamma if not given
        a = ones((self.levels,2))
        a[:,0] = arange(1,self.levels+1)
        if self.alpha0 <= 0:
            x = lstsq(a,log2(self.mean_level[1:]),rcond=None)[0]
            self.alpha = maximum(.5,-x[0])
        if self.beta0 <= 0:
            x = lstsq(a,log2(self.var_level[1:]),rcond=None)[0]
            self.beta = maximum(.5,-x[0])
        if self.gamma0 <= 0:
            x = lstsq(a,log2(self.cost_per_sample[1:]),rcond=None)[0]
            self.gamma = maximum(.5,x[0])
        self.n_total = self.n_level.sum()

    def _add_level(self):
        """ Add another level to relevant attributes. """
        self.levels += 1
        if not len(self.n_level) > self.levels:
            self.mean_level = hstack((self.mean_level, self.mean_level[-1] / 2**self.alpha))
            self.var_level = hstack((self.var_level, self.var_level[-1] / 2**self.beta))
            self.cost_per_sample = hstack((self.cost_per_sample, self.cost_per_sample[-1] * 2**self.gamma))
            self.n_level = hstack((self.n_level, 0.))
            self.sum_level = hstack((self.sum_level,zeros((2,1))))
            self.cost_level = hstack((self.cost_level, 0.))
        else:
            self.mean_level = absolute(self.sum_level[0,:self.levels+1]/self.n_level[:self.levels+1])
            self.var_level = maximum(0,self.sum_level[1,:self.levels+1]/self.n_level[:self.levels+1] - self.mean_level**2)
            self.cost_per_sample = self.cost_level[:self.levels+1]/self.n_level[:self.levels+1]


from ._true_measure import TrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution import DigitalNetB2
from numpy import *
from numpy.linalg import cholesky, slogdet, eigh
from scipy.stats import norm, multivariate_normal


class Gaussian(TrueMeasure):
    """
    Normal Measure.
    
    >>> g = Gaussian(DigitalNetB2(2,seed=7),mean=[1,2],covariance=[[9,4],[4,5]])
    >>> g.gen_samples(4)
    array([[ 4.03221089,  6.5276106 ],
           [-1.02398497, -0.48146467],
           [ 3.64364042,  0.8425137 ],
           [-0.04415595,  2.52642377]])
    >>> g
    Gaussian (TrueMeasure Object)
        mean            [1 2]
        covariance      [[9 4]
                        [4 5]]
        decomp_type     PCA
    """

    def __init__(self, sampler, mean=0., covariance=1., decomp_type='PCA'):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform 
            mean (float): mu for Normal(mu,sigma^2)
            covariance (ndarray): sigma^2 for Normal(mu,sigma^2). 
                A float or d (dimension) vector input will be extended to covariance*eye(d)
            decomp_type (str): method of decomposition either  
                "PCA" for principal component analysis or 
                "Cholesky" for cholesky decomposition.
        """
        self.parameters = ['mean', 'covariance', 'decomp_type']
        # default to transform from standard uniform
        self.domain = array([[0,1]])
        self._parse_sampler(sampler)
        self._parse_gaussian_params(mean,covariance,decomp_type)
        self.range = array([[-inf,inf]])
        super(Gaussian,self).__init__()
    
    def _parse_gaussian_params(self, mean, covariance, decomp_type):
        self.decomp_type = decomp_type.upper()
        self.mean = mean
        self.covariance = covariance
        if isscalar(mean):
            mean = tile(mean,self.d)
        if isscalar(covariance):
            covariance = covariance*eye(self.d)
        self.mu = array(mean)
        self.sigma = array(covariance)
        if self.sigma.shape==(self.d,):
            self.sigma = diag(self.sigma)
        if not (len(self.mu)==self.d and self.sigma.shape==(self.d,self.d)):
            raise DimensionError('''
                    mean must have length d and
                    covariance must be of shape d x d''')
        if self.decomp_type == 'PCA':
            evals,evecs = eigh(self.sigma) # get eigenvectors and eigenvalues for
            order = argsort(-evals)
            self.a = dot(evecs[:,order],diag(sqrt(evals[order])))
        elif self.decomp_type == 'CHOLESKY':
            self.a = cholesky(self.sigma) #Fred changed this
        else:
            raise ParameterError("decomp_type should be 'PCA' or 'Cholesky'") 
        self.mvn_scipy = multivariate_normal(mean=self.mu,cov=self.sigma, allow_singular=True)

    def _transform(self, x):
        return self.mu + norm.ppf(x)@self.a.T

    def _weight(self, t):
        return self.mvn_scipy.pdf(t)
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = Gaussian(sampler,mean=self.mu,covariance=self.covariance,decomp_type=self.decomp_type)
        else:
            m = self.mu[0]
            c = self.sigma[0,0]
            expected_cov = c*eye(int(self.d))
            if not ( (self.mu==m).all() and (self.sigma==expected_cov).all() ):
                raise DimensionError('''
                        In order to spawn a Gaussian measure
                        mean (mu) must be all the same and 
                        covariance must be a scaler times I''')
            spawn = Gaussian(sampler,mean=m,covariance=c,decomp_type=self.decomp_type)
        return spawn

class Normal(Gaussian): pass



from ._accumulate_data import AccumulateData
from ..util import MaxSamplesWarning
from ..discrete_distribution import Lattice
from numpy import array, nan
import warnings
import numpy as np
from scipy.optimize import fminbound as fminbnd
from scipy.optimize import fmin, fmin_bfgs
from numpy import sqrt, exp, log
from scipy.stats import norm as gaussnorm
from scipy.stats import t as tnorm


class LDTransformBayesData(AccumulateData):
    """
    Update and store transformation data based on low-discrepancy sequences.
    See the stopping criterion that utilize this object for references.
    """

    def __init__(self, stopping_crit, integrand, true_measure, discrete_distrib, m_min: int, m_max: int,
                 fbt, merge_fbt, kernel, alpha):
        """
        Args:
            stopping_crit (StoppingCriterion): a StoppingCriterion instance
            integrand (Integrand): an Integrand instance
            true_measure (TrueMeasure): A TrueMeasure instance
            discrete_distrib (DiscreteDistribution): a DiscreteDistribution instance
            m_min (int): initial n == 2^m_min
            m_max (int): max n == 2^m_max
        """
        self.parameters = ['solution', 'error_bound', 'n_total']
        self.stopping_crit = stopping_crit
        self.integrand = integrand
        self.true_measure = true_measure
        self.discrete_distrib = discrete_distrib
        self.distribution_name = type(self.discrete_distrib).__name__

        # Bayes cubature properties
        self.errbd_type = self.stopping_crit.errbd_type
        self.arb_mean = self.stopping_crit.arb_mean
        self.order = self.stopping_crit.order
        self.kernType = self.stopping_crit.kernType
        self.avoid_cancel_error = self.stopping_crit.avoid_cancel_error
        self.abs_tol = self.stopping_crit.abs_tol
        self.rel_tol = self.stopping_crit.rel_tol
        self.debug_enable = self.stopping_crit.debug_enable

        # Credible interval : two-sided confidence, i.e., 1-alpha percent quantile
        # quantile value for the error bound
        if self.errbd_type == 'full_Bayes':
            # degrees of freedom = 2^mmin - 1
            self.uncert = -tnorm.ppf(alpha / 2, (2 ** m_min) - 1)
        else:
            self.uncert = -gaussnorm.ppf(alpha / 2)

        # Set Attributes
        self.m_min = m_min
        self.m_max = m_max
        self.debugEnable = True

        self.n_total = 0  # total number of samples generated
        self.solution = nan

        self.iter = 0
        self.m = self.m_min
        self.mvec = np.arange(self.m_min, self.m_max + 1, dtype=int)

        # Initialize various temporary storage between iterations
        self.xpts_ = array([])  # shifted lattice points
        self.xun_ = array([])  # un-shifted lattice points
        self.ftilde_ = array([])  # fourier transformed integrand values
        if isinstance(self.discrete_distrib, Lattice):
            # integrand after the periodization transform
            self.ff = lambda x, *args, **kwargs: self.integrand.f(x,
                                                                  periodization_transform=stopping_crit.ptransform,
                                                                  *args, **kwargs).squeeze()
        else:
            self.ff = self.integrand.f
        self.fbt = fbt
        self.merge_fbt = merge_fbt
        self.kernel = kernel

        super(LDTransformBayesData, self).__init__()

    def update_data(self, y_val_new=None, xnew=None, xunnew=None):
        """ See abstract method. """
        # Generate sample values

        if self.iter < len(self.mvec):
            if y_val_new is None:
                self.ftilde_, self.xun_, self.xpts_ = self.iter_fbt(self.iter, self.xun_, self.xpts_, self.ftilde_)
            else:
                self.ftilde_, self.xun_, self.xpts_ = self.iter_fbt(self.iter, self.xun_, self.xpts_, self.ftilde_,
                                                                    y_val_new, xunnew, xnew)
            self.m = self.mvec[self.iter]
            self.iter += 1
            # update total samples
            self.n_total = 2 ** self.m  # updated the total evaluations
        else:
            warnings.warn('''
                Already used maximum allowed sample size %d.
                Note that error tolerances may no longer be satisfied.''' % (2 ** self.m_max),
                          MaxSamplesWarning)

        return self._stopping_criterion(self.xun_, self.ftilde_, self.m)

    # decides if the user-defined error threshold is met
    def _stopping_criterion(self, xpts, ftilde, m):
        r = self.stopping_crit.order
        ftilde = ftilde.squeeze()
        n = 2 ** m
        success = False
        lna_range = [-5, 0]  # reduced from [-5, 5], to avoid kernel values getting too big causing error

        # search for optimal shape parameter
        if self.stopping_crit.one_theta == True:
            lna_MLE = fminbnd(lambda lna: self.objective_function(exp(lna), xpts, ftilde)[0],
                              x1=lna_range[0], x2=lna_range[1], xtol=1e-2, disp=0)

            aMLE = exp(lna_MLE)
            _, vec_lambda, vec_lambda_ring, RKHS_norm = self.objective_function(aMLE, xpts, ftilde)
        else:
            if self.stopping_crit.use_gradient == True:
                warnings.warn('Not implemented !')
                lna_MLE = 0
            else:
                # Nelder-Mead Simplex algorithm
                theta0 = np.ones((xpts.shape[1], 1)) * (0.05)
                theta0 = np.ones((1, xpts.shape[1])) * (0.05)
                lna_MLE = fmin(lambda lna: self.objective_function(exp(lna), xpts, ftilde)[0],
                               theta0, xtol=1e-2, disp=False)
            aMLE = exp(lna_MLE)
            # print(n, aMLE)
            _, vec_lambda, vec_lambda_ring, RKHS_norm = self.objective_function(aMLE, xpts, ftilde)

        # Check error criterion
        # compute DSC
        if self.errbd_type == 'full_Bayes':
            # full Bayes
            if self.avoid_cancel_error:
                DSC = abs(vec_lambda_ring[0] / n)
            else:
                DSC = abs((vec_lambda[0] / n) - 1)

            # 1-alpha two sided confidence interval
            err_bd = self.uncert * sqrt(DSC * RKHS_norm / (n - 1))
        elif self.errbd_type == 'GCV':
            # GCV based stopping criterion
            if self.avoid_cancel_error:
                DSC = abs(vec_lambda_ring[0] / (n + vec_lambda_ring[0]))
            else:
                DSC = abs(1 - (n / vec_lambda[0]))

            temp = vec_lambda
            temp[0] = n + vec_lambda_ring[0]
            mC_inv_trace = sum(1. / temp(temp != 0))
            err_bd = self.uncert * sqrt(DSC * RKHS_norm / mC_inv_trace)
        else:
            # empirical Bayes
            if self.avoid_cancel_error:
                DSC = abs(vec_lambda_ring[0] / (n + vec_lambda_ring[0]))
            else:
                DSC = abs(1 - (n / vec_lambda[0]))
            err_bd = self.uncert * sqrt(DSC * RKHS_norm / n)

        if self.arb_mean:  # zero mean case
            muhat = ftilde[0] / n
        else:  # non zero mean case
            muhat = ftilde[0] / vec_lambda[0]

        self.error_bound = err_bd
        muhat = np.abs(muhat)
        muminus = muhat - err_bd
        muplus = muhat + err_bd

        if 2 * err_bd <= max(self.abs_tol, self.rel_tol * abs(muminus)) + max(self.abs_tol, self.rel_tol * abs(muplus)):
            if err_bd == 0:
                err_bd = np.finfo(float).eps

            # stopping criterion achieved
            success = True
        return success, muhat, r, err_bd, m

    # objective function to estimate parameter theta
    # MLE : Maximum likelihood estimation
    # GCV : Generalized cross validation
    def objective_function(self, theta, xun, ftilde):
        n = len(ftilde)
        fudge = 100 * np.finfo(float).eps
        # if type(theta) != np.ndarray:
        #     theta = np.ones((1, xun.shape[1])) * theta
        [vec_lambda, vec_lambda_ring, lambda_factor] = self.kernel(xun, self.order, theta, self.avoid_cancel_error,
                                                                   self.kernType, self.debug_enable)
        vec_lambda = abs(vec_lambda)
        # compute RKHS_norm
        temp = abs(ftilde[vec_lambda > fudge] ** 2) / (vec_lambda[vec_lambda > fudge])

        # compute loss
        if self.errbd_type == 'GCV':
            # GCV
            temp_gcv = abs(ftilde[vec_lambda > fudge] / (vec_lambda[vec_lambda > fudge])) ** 2
            loss1 = 2 * log(sum(1. / vec_lambda[vec_lambda > fudge]))
            loss2 = log(sum(temp_gcv[1:]))
            # ignore all zero eigenvalues
            loss = loss2 - loss1

            if self.arb_mean:
                RKHS_norm = (1 / lambda_factor) * sum(temp_gcv[1:]) / n
            else:
                RKHS_norm = (1 / lambda_factor) * sum(temp_gcv) / n
        else:
            # default: MLE
            if self.arb_mean:
                RKHS_norm = (1 / lambda_factor) * sum(temp[1:]) / n
                temp_1 = (1 / lambda_factor) * sum(temp[1:])
            else:
                RKHS_norm = (1 / lambda_factor) * sum(temp) / n
                temp_1 = (1 / lambda_factor) * sum(temp)

            # ignore all zero eigenvalues
            loss1 = sum(log(abs(lambda_factor * vec_lambda[vec_lambda > fudge])))
            if temp_1 != 0:
                loss2 = n * log(temp_1)
            else:
                loss2 = n * log(temp_1 + np.finfo(float).eps)
            loss = loss1 + loss2

        if self.debug_enable:
            self.alert_msg(loss1, 'Inf', 'Imag')
            self.alert_msg(RKHS_norm, 'Imag')
            self.alert_msg(loss2, 'Inf', 'Imag')
            self.alert_msg(loss, 'Inf', 'Imag', 'Nan')
            self.alert_msg(vec_lambda, 'Imag')

        vec_lambda, vec_lambda_ring = lambda_factor * vec_lambda, lambda_factor * vec_lambda_ring
        return loss, vec_lambda, vec_lambda_ring, RKHS_norm

    # Efficient Fast Bayesian Transform computation algorithm, avoids recomputing the full transform
    def iter_fbt(self, iter, xun_, xpts_, ftilde_prev, y_val_new=None, xunnew=None, xnew=None):
        m = self.mvec[iter]
        n = 2 ** m

        # In every iteration except the first one, "n" number_of_points is doubled,
        # but FBT is only computed for the newly added points.
        # Previously computed FFT is reused.
        if iter == 0:
            # In the first iteration compute full FBT
            # xun_ = mod(bsxfun( @ times, (0:1 / n:1-1 / n)',self.gen_vec),1)
            # xun_ = np.arange(0, 1, 1 / n).reshape((n, 1))
            # xun_ = np.mod((xun_ * self.gen_vec), 1)
            # xpts_ = np.mod(bsxfun( @ plus, xun_, shift), 1)  # shifted

            if xnew is None:
                xpts_, xun_ = self.gen_samples(n_min=0, n_max=n, return_unrandomized=True,
                                               distribution=self.discrete_distrib)

                # Compute initial FBT
                ftilde_ = self.fbt(self.ff(xpts_))
            else:
                xpts_, xun_ = xnew, xunnew
                ftilde_ = self.fbt(y_val_new)

            ftilde_ = ftilde_.reshape((n, 1))
        else:
            # xunnew = np.mod(bsxfun( @ times, (1/n : 2/n : 1-1/n)',self.gen_vec),1)
            # xunnew = np.arange(1 / n, 1, 2 / n).reshape((n // 2, 1))
            # xunnew = np.mod(xunnew * self.gen_vec, 1)
            # xnew = np.mod(bsxfun( @ plus, xunnew, shift), 1)

            if xnew is None:
                xnew, xunnew = self.gen_samples(n_min=n // 2, n_max=n, return_unrandomized=True,
                                                distribution=self.discrete_distrib)
            [xun_, xpts_] = self.merge_pts(xun_, xunnew, xpts_, xnew, n, self.discrete_distrib.d,
                                           distribution=self.distribution_name)
            mnext = m - 1
            if y_val_new is None:
                ftilde_next_new = self.fbt(self.ff(xnew))
            else:
                ftilde_next_new = self.fbt(y_val_new)

            ftilde_next_new = ftilde_next_new.reshape((n // 2, 1))
            if self.debugEnable:
                self.alert_msg(ftilde_next_new, 'Nan', 'Inf')

            # combine the previous batch and new batch to get FBT on all points
            ftilde_ = self.merge_fbt(ftilde_prev, ftilde_next_new, mnext)

        return ftilde_, xun_, xpts_

    @staticmethod
    def gen_samples(n_min, n_max, return_unrandomized, distribution):
        warn = False
        xpts_, xun_ = distribution.gen_samples(n_min=n_min, n_max=n_max, warn=warn,
                                               return_unrandomized=return_unrandomized)
        return xpts_, xun_

    # inserts newly generated points with the old set by interleaving them
    # xun - unshifted points
    @staticmethod
    def merge_pts(xun, xunnew, x, xnew, n, d, distribution):
        if distribution == 'Lattice':
            temp = np.zeros((n, d))
            temp[0::2, :] = xun
            try:
                temp[1::2, :] = xunnew
            except Exception as e:
                raise(e)
            xun = temp
            temp = np.zeros((n, d))
            temp[0::2, :] = x
            temp[1::2, :] = xnew
            x = temp
        else:
            x = np.vstack([x, xnew])
            xun = np.vstack([xun, xunnew])
        return xun, x

    # Computes modified kernel Km1 = K - 1
    # Useful to avoid cancellation error in the computation of (1 - n/\lambda_1)
    @staticmethod
    def kernel_t(aconst, Bern):
        d = np.size(Bern, 1)
        if type(aconst) != np.ndarray:
            theta = np.ones((d, 1)) * aconst
        else:
            theta = aconst  # theta varies per dimension

        Kjm1 = theta[0] * Bern[:, 0]  # Kernel at j-dim minus One
        Kj = 1 + Kjm1  # Kernel at j-dim

        for j in range(1, d):
            Kjm1_prev = Kjm1
            Kj_prev = Kj  # save the Kernel at the prev dim

            Kjm1 = theta[j] * Bern[:, j] * Kj_prev + Kjm1_prev
            Kj = 1 + Kjm1

        Km1 = Kjm1
        K = Kj
        return [Km1, K]

    # prints debug message if the given variable is Inf, Nan or complex, etc
    # Example: alertMsg(x, 'Inf', 'Imag')
    #          prints if variable 'x' is either Infinite or Imaginary
    @staticmethod
    def alert_msg(*args):
        varargin = args
        nargin = len(varargin)
        if nargin > 1:
            i_start = 0
            var_tocheck = varargin[i_start]
            i_start = i_start + 1
            inpvarname = 'variable'

            while i_start < nargin:
                var_type = varargin[i_start]
                i_start = i_start + 1

                if var_type == 'Nan':
                    if np.any(np.isnan(var_tocheck)):
                        print('%s has NaN values' % inpvarname)
                elif var_type == 'Inf':
                    if np.any(np.isinf(var_tocheck)):
                        print('%s has Inf values' % inpvarname)
                elif var_type == 'Imag':
                    if not np.all(np.isreal(var_tocheck)):
                        print('%s has complex values' % inpvarname)
                else:
                    print('unknown type check requested !')
